"""

MSSP - Mud Server Status Protocol

This implements the MSSP telnet protocol as per
http://tintin.sourceforge.net/mssp/.  MSSP allows web portals and
listings to have their crawlers find the mud and automatically
extract relevant information about it, such as genre, how many
active players and so on.


"""
from django.conf import settings

from evennia.utils import utils

MSSP = bytes([70])  # b"\x46"
MSSP_VAR = bytes([1])  # b"\x01"
MSSP_VAL = bytes([2])  # b"\x02"

# try to get the customized mssp info, if it exists.
MSSPTable_CUSTOM = utils.variable_from_module(settings.MSSP_META_MODULE, "MSSPTable", default={})


class Mssp:
    """
    Implements the MSSP protocol. Add this to a variable on the telnet
    protocol to set it up.

    """

    def __init__(self, protocol):
        """
        initialize MSSP by storing protocol on ourselves and calling
        the client to see if it supports MSSP.

        Args:
            protocol (Protocol): The active protocol instance.

        """
        self.protocol = protocol
        self.protocol.will(MSSP).addCallbacks(self.do_mssp, self.no_mssp)

    def get_player_count(self):
        """
        Get number of logged-in players.

        Returns:
            count (int): The number of players in the MUD.

        """
        return str(self.protocol.sessionhandler.count_loggedin())

    def get_uptime(self):
        """
        Get how long the portal has been online (reloads are not counted).

        Returns:
            uptime (int): Number of seconds of uptime.

        """
        return str(self.protocol.sessionhandler.uptime)

    def no_mssp(self, option):
        """
        Called when mssp is not requested. This is the normal
        operation.

        Args:
            option (Option): Not used.

        """
        self.protocol.handshake_done()

    def do_mssp(self, option):
        """
        Negotiate all the information.

        Args:
            option (Option): Not used.

        """

        self.mssp_table = {
            # Required fields
            "NAME": settings.SERVERNAME,
            "PLAYERS": self.get_player_count,
            "UPTIME": self.get_uptime,
            "PORT": list(
                str(port) for port in reversed(settings.TELNET_PORTS)
            ),  # most important port should be last in list
            # Evennia auto-filled
            "CRAWL DELAY": "-1",
            "CODEBASE": utils.get_evennia_version(mode="pretty"),
            "FAMILY": "Custom",
            "ANSI": "1",
            "GMCP": "1" if settings.TELNET_OOB_ENABLED else "0",
            "ATCP": "0",
            "MCCP": "1",
            "MCP": "0",
            "MSDP": "1" if settings.TELNET_OOB_ENABLED else "0",
            "MSP": "0",
            "MXP": "1",
            "PUEBLO": "0",
            "SSL": "1" if settings.SSL_ENABLED else "0",
            "UTF-8": "1",
            "ZMP": "0",
            "VT100": "1",
            "XTERM 256 COLORS": "1",
        }

        # update the static table with the custom one
        if MSSPTable_CUSTOM:
            self.mssp_table.update(MSSPTable_CUSTOM)

        varlist = b""
        for variable, value in self.mssp_table.items():
            if callable(value):
                value = value()
            if utils.is_iter(value):
                for partval in value:
                    varlist += (
                        MSSP_VAR
                        + bytes(str(variable), "utf-8")
                        + MSSP_VAL
                        + bytes(str(partval), "utf-8")
                    )
            else:
                varlist += (
                    MSSP_VAR + bytes(str(variable), "utf-8") + MSSP_VAL + bytes(str(value), "utf-8")
                )

        # send to crawler by subnegotiation
        self.protocol.requestNegotiation(MSSP, varlist)
        self.protocol.handshake_done()



"""
Simple turn-based combat system with range and movement

Contrib - Tim Ashley Jenkins 2017

This is a version of the 'turnbattle' contrib that includes a system
for abstract movement and positioning in combat, including distinction
between melee and ranged attacks. In this system, a fighter or object's
exact position is not recorded - only their relative distance to other
actors in combat.

In this example, the distance between two objects in combat is expressed
as an integer value: 0 for "engaged" objects that are right next to each
other, 1 for "reach" which is for objects that are near each other but
not directly adjacent, and 2 for "range" for objects that are far apart.

When combat starts, all fighters are at reach with each other and other
objects, and at range from any exits. On a fighter's turn, they can use
the "approach" command to move closer to an object, or the "withdraw"
command to move further away from an object, either of which takes an
action in combat. In this example, fighters are given two actions per
turn, allowing them to move and attack in the same round, or to attack
twice or move twice.

When you move toward an object, you will also move toward anything else
that's close to your target - the same goes for moving away from a target,
which will also move you away from anything close to your target. Moving
toward one target may also move you away from anything you're already
close to, but withdrawing from a target will never inadvertently bring
you closer to anything else.

In this example, there are two attack commands. 'Attack' can only hit
targets that are 'engaged' (range 0) with you. 'Shoot' can hit any target
on the field, but cannot be used if you are engaged with any other fighters.
In addition, strikes made with the 'attack' command are more accurate than
'shoot' attacks. This is only to provide an example of how melee and ranged
attacks can be made to work differently - you can, of course, modify this
to fit your rules system.

When in combat, the ranges of objects are also accounted for - you can't
pick up an object unless you're engaged with it, and can't give an object
to another fighter without being engaged with them either. Dropped objects
are automatically assigned a range of 'engaged' with the fighter who dropped
them. Additionally, giving or getting an object will take an action in combat.
Dropping an object does not take an action, but can only be done on your turn.

When combat ends, all range values are erased and all restrictions on getting
or getting objects are lifted - distances are no longer tracked and objects in
the same room can be considered to be in the same space, as is the default
behavior of Evennia and most MUDs.

This system allows for strategies in combat involving movement and
positioning to be implemented in your battle system without the use of
a 'grid' of coordinates, which can be difficult and clunky to navigate
in text and disadvantageous to players who use screen readers. This loose,
narrative method of tracking position is based around how the matter is
handled in tabletop RPGs played without a grid - typically, a character's
exact position in a room isn't important, only their relative distance to
other actors.

You may wish to expand this system with a method of distinguishing allies
from enemies (to prevent allied characters from blocking your ranged attacks)
as well as some method by which melee-focused characters can prevent enemies
from withdrawing or punish them from doing so, such as by granting "attacks of
opportunity" or something similar. If you wish, you can also expand the breadth
of values allowed for range - rather than just 0, 1, and 2, you can allow ranges
to go up to much higher values, and give attacks and movements more varying
values for distance for a more granular system. You may also want to implement
a system for fleeing or changing rooms in combat by approaching exits, which
are objects placed in the range field like any other.

To install and test, import this module's TBRangeCharacter object into
your game's character.py module:

    from evennia.contrib.game_systems.turnbattle.tb_range import TBRangeCharacter

And change your game's character typeclass to inherit from TBRangeCharacter
instead of the default:

    class Character(TBRangeCharacter):

Do the same thing in your game's objects.py module for TBRangeObject:

    from evennia.contrib.game_systems.turnbattle.tb_range import TBRangeObject
    class Object(TBRangeObject):

Next, import this module into your default_cmdsets.py module:

    from evennia.contrib.game_systems.turnbattle import tb_range

And add the battle command set to your default command set:

    #
    # any commands you add below will overload the default ones.
    #
    self.add(tb_range.BattleCmdSet())

This module is meant to be heavily expanded on, so you may want to copy it
to your game's 'world' folder and modify it there rather than importing it
in your game and using it as-is.
"""

from random import randint

from evennia import Command, DefaultObject, DefaultScript, default_cmds
from evennia.commands.default.help import CmdHelp

from . import tb_basic

"""
----------------------------------------------------------------------------
OPTIONS
----------------------------------------------------------------------------
"""

TURN_TIMEOUT = 30  # Time before turns automatically end, in seconds
ACTIONS_PER_TURN = 2  # Number of actions allowed per turn

"""
----------------------------------------------------------------------------
COMBAT FUNCTIONS START HERE
----------------------------------------------------------------------------
"""


class RangedCombatRules(tb_basic.BasicCombatRules):
    def get_attack(self, attacker, defender, attack_type):
        """
        Returns a value for an attack roll.

        Args:
            attacker (obj): Character doing the attacking
            defender (obj): Character being attacked
            attack_type (str): Type of attack ('melee' or 'ranged')

        Returns:
            attack_value (int): Attack roll value, compared against a defense value
                to determine whether an attack hits or misses.

        Notes:
            By default, generates a random integer from 1 to 100 without using any
            properties from either the attacker or defender, and modifies the result
            based on whether it's for a melee or ranged attack.

            This can easily be expanded to return a value based on characters stats,
            equipment, and abilities. This is why the attacker and defender are passed
            to this function, even though nothing from either one are used in this example.
        """
        # For this example, just return a random integer up to 100.
        attack_value = randint(1, 100)
        # Make melee attacks more accurate, ranged attacks less accurate
        if attack_type == "melee":
            attack_value += 15
        if attack_type == "ranged":
            attack_value -= 15
        return attack_value

    def get_defense(self, attacker, defender, attack_type="melee"):
        """
        Returns a value for defense, which an attack roll must equal or exceed in order
        for an attack to hit.

        Args:
            attacker (obj): Character doing the attacking
            defender (obj): Character being attacked
            attack_type (str): Type of attack ('melee' or 'ranged')

        Returns:
            defense_value (int): Defense value, compared against an attack roll
                to determine whether an attack hits or misses.

        Notes:
            By default, returns 50, not taking any properties of the defender or
            attacker into account.

            As above, this can be expanded upon based on character stats and equipment.
        """
        # For this example, just return 50, for about a 50/50 chance of hit.
        defense_value = 50
        return defense_value

    def get_range(self, obj1, obj2):
        """
        Gets the combat range between two objects.

        Args:
            obj1 (obj): First object
            obj2 (obj): Second object

        Returns:
            range (int or None): Distance between two objects or None if not applicable
        """
        # Return None if not applicable.
        if not obj1.db.combat_range:
            return None
        if not obj2.db.combat_range:
            return None
        if obj1 not in obj2.db.combat_range:
            return None
        if obj2 not in obj1.db.combat_range:
            return None
        # Return the range between the two objects.
        return obj1.db.combat_range[obj2]

    def distance_inc(self, mover, target):
        """
        Function that increases distance in range field between mover and target.

        Args:
            mover (obj): The object moving
            target (obj): The object to be moved away from
        """
        mover.db.combat_range[target] += 1
        target.db.combat_range[mover] = mover.db.combat_range[target]
        # Set a cap of 2:
        if self.get_range(mover, target) > 2:
            target.db.combat_range[mover] = 2
            mover.db.combat_range[target] = 2

    def distance_dec(self, mover, target):
        """
        Helper function that decreases distance in range field between mover and target.

        Args:
            mover (obj): The object moving
            target (obj): The object to be moved toward
        """
        mover.db.combat_range[target] -= 1
        target.db.combat_range[mover] = mover.db.combat_range[target]
        # If this brings mover to range 0 (Engaged):
        if self.get_range(mover, target) <= 0:
            # Reset range to each other to 0 and copy target's ranges to mover.
            target.db.combat_range[mover] = 0
            mover.db.combat_range = target.db.combat_range
            # Assure everything else has the same distance from the mover and target, now that
            # they're together
            for thing in mover.location.contents:
                if thing != mover and thing != target:
                    thing.db.combat_range[mover] = thing.db.combat_range[target]

    def approach(self, mover, target):
        """
        Manages a character's whole approach, including changes in ranges to other characters.

        Args:
            mover (obj): The object moving
            target (obj): The object to be moved toward

        Notes:
            The mover will also automatically move toward any objects that are closer to the
            target than the mover is. The mover will also move away from anything they started
            out close to.
        """

        contents = mover.location.contents

        for thing in contents:
            if thing != mover and thing != target:
                # Move closer to each object closer to the target than you.
                if self.get_range(mover, thing) > self.get_range(target, thing):
                    self.distance_dec(mover, thing)
                # Move further from each object that's further from you than from the target.
                if self.get_range(mover, thing) < self.get_range(target, thing):
                    self.distance_inc(mover, thing)
        # Lastly, move closer to your target.
        self.distance_dec(mover, target)

    def withdraw(self, mover, target):
        """
        Manages a character's whole withdrawal, including changes in ranges to other characters.

        Args:
            mover (obj): The object moving
            target (obj): The object to be moved away from

        Notes:
            The mover will also automatically move away from objects that are close to the target
            of their withdrawl. The mover will never inadvertently move toward anything else while
            withdrawing - they can be considered to be moving to open space.
        """

        contents = mover.location.contents

        for thing in contents:
            if thing != mover and thing != target:
                # Move away from each object closer to the target than you, if it's also closer to
                # you than you are to the target.
                if self.get_range(mover, thing) >= self.get_range(target, thing) and self.get_range(
                    mover, thing
                ) < self.get_range(mover, target):
                    self.distance_inc(mover, thing)
                # Move away from anything your target is engaged with
                if self.get_range(target, thing) == 0:
                    self.distance_inc(mover, thing)
                # Move away from anything you're engaged with.
                if self.get_range(mover, thing) == 0:
                    self.distance_inc(mover, thing)
        # Then, move away from your target.
        self.distance_inc(mover, target)

    def resolve_attack(
        self, attacker, defender, attack_value=None, defense_value=None, attack_type="melee"
    ):
        """
        Resolves an attack and outputs the result.

        Args:
            attacker (obj): Character doing the attacking
            defender (obj): Character being attacked
            attack_type (str): Type of attack (melee or ranged)

        Notes:
            Even though the attack and defense values are calculated
            extremely simply, they are separated out into their own functions
            so that they are easier to expand upon.

        """
        # Get an attack roll from the attacker.
        if not attack_value:
            attack_value = self.get_attack(attacker, defender, attack_type)
        # Get a defense value from the defender.
        if not defense_value:
            defense_value = self.get_defense(attacker, defender, attack_type)
        # If the attack value is lower than the defense value, miss. Otherwise, hit.
        if attack_value < defense_value:
            attacker.location.msg_contents(
                "%s's %s attack misses %s!" % (attacker, attack_type, defender)
            )
        else:
            damage_value = self.get_damage(attacker, defender)  # Calculate damage value.
            # Announce damage dealt and apply damage.
            attacker.location.msg_contents(
                "%s hits %s with a %s attack for %i damage!"
                % (attacker, defender, attack_type, damage_value)
            )
            self.apply_damage(defender, damage_value)
            # If defender HP is reduced to 0 or less, call at_defeat.
            if defender.db.hp <= 0:
                self.at_defeat(defender)

    def combat_status_message(self, fighter):
        """
        Sends a message to a player with their current HP and
        distances to other fighters and objects. Called at turn
        start and by the 'status' command.
        """
        if not fighter.db.max_hp:
            fighter.db.hp = 100
            fighter.db.max_hp = 100

        status_msg = "HP Remaining: %i / %i" % (fighter.db.hp, fighter.db.max_hp)

        if not self.is_in_combat(fighter):
            fighter.msg(status_msg)
            return

        engaged_obj = []
        reach_obj = []
        range_obj = []

        for thing in fighter.db.combat_range:
            if thing != fighter:
                if fighter.db.combat_range[thing] == 0:
                    engaged_obj.append(thing)
                if fighter.db.combat_range[thing] == 1:
                    reach_obj.append(thing)
                if fighter.db.combat_range[thing] > 1:
                    range_obj.append(thing)

        if engaged_obj:
            status_msg += "|/Engaged targets: %s" % ", ".join(obj.key for obj in engaged_obj)
        if reach_obj:
            status_msg += "|/Reach targets: %s" % ", ".join(obj.key for obj in reach_obj)
        if range_obj:
            status_msg += "|/Ranged targets: %s" % ", ".join(obj.key for obj in range_obj)

        fighter.msg(status_msg)


COMBAT_RULES = RangedCombatRules()

"""
----------------------------------------------------------------------------
SCRIPTS START HERE
----------------------------------------------------------------------------
"""


class TBRangeTurnHandler(tb_basic.TBBasicTurnHandler):
    """
    This is the script that handles the progression of combat through turns.
    On creation (when a fight is started) it adds all combat-ready characters
    to its roster and then sorts them into a turn order. There can only be one
    fight going on in a single room at a time, so the script is assigned to a
    room as its object.

    Fights persist until only one participant is left with any HP or all
    remaining participants choose to end the combat with the 'disengage'
    command.
    """

    rules = COMBAT_RULES

    def init_range(self, to_init):
        """
        Initializes range values for an object at the start of a fight.

        Args:
            to_init (object): Object to initialize range field for.
        """
        rangedict = {}
        # Get a list of objects in the room.
        objectlist = self.obj.contents
        for thing in objectlist:
            # Object always at distance 0 from itself
            if thing == to_init:
                rangedict.update({thing: 0})
            else:
                if thing.destination or to_init.destination:
                    # Start exits at range 2 to put them at the 'edges'
                    rangedict.update({thing: 2})
                else:
                    # Start objects at range 1 from other objects
                    rangedict.update({thing: 1})
        to_init.db.combat_range = rangedict

    def join_rangefield(self, to_init, anchor_obj=None, add_distance=0):
        """
        Adds a new object to the range field of a fight in progress.

        Args:
            to_init (object): Object to initialize range field for.
        Keyword Args:
            anchor_obj (object): Object to copy range values from, or None for a random object.
            add_distance (int): Distance to put between to_init object and anchor object.

        """
        # Get a list of room's contents without to_init object.
        contents = self.obj.contents
        contents.remove(to_init)
        # If no anchor object given, pick one in the room at random.
        if not anchor_obj:
            anchor_obj = contents[randint(0, (len(contents) - 1))]
        # Copy the range values from the anchor object.
        to_init.db.combat_range = anchor_obj.db.combat_range
        # Add the new object to everyone else's ranges.
        for thing in contents:
            new_objects_range = thing.db.combat_range[anchor_obj]
            thing.db.combat_range.update({to_init: new_objects_range})
        # Set the new object's range to itself to 0.
        to_init.db.combat_range.update({to_init: 0})
        # Add additional distance from anchor object, if any.
        for n in range(add_distance):
            self.rules.withdraw(to_init, anchor_obj)

    def start_turn(self, character):
        """
        Readies a character for the start of their turn by replenishing their
        available actions and notifying them that their turn has come up.

        Args:
            character (obj): Character to be readied.

        Notes:
            In this example, characters are given two actions per turn. This allows
            characters to both move and attack in the same turn (or, alternately,
            move twice or attack twice).
        """
        super().start_turn(character)
        character.db.combat_actionsleft = ACTIONS_PER_TURN

    def join_fight(self, character):
        """
        Adds a new character to a fight already in progress.

        Args:
            character (obj): Character to be added to the fight.
        """
        # Inserts the fighter to the turn order, right behind whoever's turn it currently is.
        self.db.fighters.insert(self.db.turn, character)
        # Tick the turn counter forward one to compensate.
        self.db.turn += 1
        # Initialize the character like you do at the start.
        self.initialize_for_combat(character)
        # Add the character to the rangefield, at range from everyone, if they're not on it already.
        if not character.db.combat_range:
            self.join_rangefield(character, add_distance=2)


"""
----------------------------------------------------------------------------
TYPECLASSES START HERE
----------------------------------------------------------------------------
"""


class TBRangeCharacter(tb_basic.TBBasicCharacter):
    """
    A character able to participate in turn-based combat. Has attributes for current
    and maximum HP, and access to combat commands.
    """

    rules = COMBAT_RULES


class TBRangeObject(DefaultObject):
    """
    An object that is assigned range values in combat. Getting, giving, and dropping
    the object has restrictions in combat - you must be next to an object to get it,
    must be next to your target to give them something, and can only interact with
    objects on your own turn.
    """

    def at_pre_drop(self, dropper):
        """
        Called by the default `drop` command before this object has been
        dropped.

        Args:
            dropper (Object): The object which will drop this object.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Returns:
            shoulddrop (bool): If the object should be dropped or not.

        Notes:
            If this method returns False/None, the dropping is cancelled
            before it is even started.

        """
        # Can't drop something if in combat and it's not your turn
        if self.rules.is_in_combat(dropper) and not self.rules.is_turn(dropper):
            dropper.msg("You can only drop things on your turn!")
            return False
        return True

    def at_drop(self, dropper):
        """
        Called by the default `drop` command when this object has been
        dropped.

        Args:
            dropper (Object): The object which just dropped this object.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Notes:
            This hook cannot stop the drop from happening. Use
            permissions or the at_pre_drop() hook for that.

        """
        # If dropper is currently in combat
        if dropper.location.db.combat_turnhandler:
            # Object joins the range field
            self.db.combat_range = {}
            dropper.location.db.combat_turnhandler.join_rangefield(self, anchor_obj=dropper)

    def at_pre_get(self, getter):
        """
        Called by the default `get` command before this object has been
        picked up.

        Args:
            getter (Object): The object about to get this object.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Returns:
            shouldget (bool): If the object should be gotten or not.

        Notes:
            If this method returns False/None, the getting is cancelled
            before it is even started.
        """
        # Restrictions for getting in combat
        if self.rules.is_in_combat(getter):
            if not self.rules.is_turn(getter):  # Not your turn
                getter.msg("You can only get things on your turn!")
                return False
            if self.rules.get_range(self, getter) > 0:  # Too far away
                getter.msg("You aren't close enough to get that! (see: help approach)")
                return False
        return True

    def at_get(self, getter):
        """
        Called by the default `get` command when this object has been
        picked up.

        Args:
            getter (Object): The object getting this object.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Notes:
            This hook cannot stop the pickup from happening. Use
            permissions or the at_pre_get() hook for that.

        """
        # If gotten, erase range values
        if self.db.combat_range:
            del self.db.combat_range
        # Remove this object from everyone's range fields
        for thing in getter.location.contents:
            if thing.db.combat_range:
                if self in thing.db.combat_range:
                    thing.db.combat_range.pop(self, None)
        # If in combat, getter spends an action
        if self.rules.is_in_combat(getter):
            self.rules.spend_action(getter, 1, action_name="get")  # Use up one action.

    def at_pre_give(self, giver, getter):
        """
        Called by the default `give` command before this object has been
        given.

        Args:
            giver (Object): The object about to give this object.
            getter (Object): The object about to get this object.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Returns:
            shouldgive (bool): If the object should be given or not.

        Notes:
            If this method returns False/None, the giving is cancelled
            before it is even started.

        """
        # Restrictions for giving in combat
        if self.rules.is_in_combat(giver):
            if not self.rules.is_turn(giver):  # Not your turn
                giver.msg("You can only give things on your turn!")
                return False
            if self.rules.get_range(giver, getter) > 0:  # Too far away from target
                giver.msg(
                    "You aren't close enough to give things to %s! (see: help approach)" % getter
                )
                return False
        return True

    def at_give(self, giver, getter):
        """
        Called by the default `give` command when this object has been
        given.

        Args:
            giver (Object): The object giving this object.
            getter (Object): The object getting this object.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Notes:
            This hook cannot stop the give from happening. Use
            permissions or the at_pre_give() hook for that.

        """
        # Spend an action if in combat
        if self.rules.is_in_combat(giver):
            self.rules.spend_action(giver, 1, action_name="give")  # Use up one action.


"""
----------------------------------------------------------------------------
COMMANDS START HERE
----------------------------------------------------------------------------
"""


class CmdFight(tb_basic.CmdFight):
    """
    Starts a fight with everyone in the same room as you.

    Usage:
      fight

    When you start a fight, everyone in the room who is able to
    fight is added to combat, and a turn order is randomly rolled.
    When it's your turn, you can attack other characters.
    """

    key = "fight"
    help_category = "combat"

    rules = COMBAT_RULES
    combat_handler_class = TBRangeTurnHandler


class CmdAttack(tb_basic.CmdAttack):
    """
    Attacks another character in melee.

    Usage:
      attack <target>

    When in a fight, you may attack another character. The attack has
    a chance to hit, and if successful, will deal damage. You can only
    attack engaged targets - that is, targets that are right next to
    you. Use the 'approach' command to get closer to a target.
    """

    key = "attack"
    help_category = "combat"

    rules = COMBAT_RULES

    def func(self):
        "This performs the actual command."
        "Set the attacker to the caller and the defender to the target."

        if not self.rules.is_in_combat(self.caller):  # If not in combat, can't attack.
            self.caller.msg("You can only do that in combat. (see: help fight)")
            return

        if not self.rules.is_turn(self.caller):  # If it's not your turn, can't attack.
            self.caller.msg("You can only do that on your turn.")
            return

        if not self.caller.db.hp:  # Can't attack if you have no HP.
            self.caller.msg("You can't attack, you've been defeated.")
            return

        attacker = self.caller
        defender = self.caller.search(self.args)

        if not defender:  # No valid target given.
            return

        if not defender.db.hp:  # Target object has no HP left or to begin with
            self.caller.msg("You can't fight that!")
            return

        if attacker == defender:  # Target and attacker are the same
            self.caller.msg("You can't attack yourself!")
            return

        if not self.rules.get_range(attacker, defender) == 0:  # Target isn't in melee
            self.caller.msg(
                "%s is too far away to attack - you need to get closer! (see: help approach)"
                % defender
            )
            return

        "If everything checks out, call the attack resolving function."
        self.rules.resolve_attack(attacker, defender, "melee")
        self.rules.spend_action(self.caller, 1, action_name="attack")  # Use up one action.


class CmdShoot(Command):
    """
    Attacks another character from range.

    Usage:
      shoot <target>

    When in a fight, you may shoot another character. The attack has
    a chance to hit, and if successful, will deal damage. You can attack
    any target in combat by shooting, but can't shoot if there are any
    targets engaged with you. Use the 'withdraw' command to retreat from
    nearby enemies.
    """

    key = "shoot"
    help_category = "combat"

    rules = COMBAT_RULES

    def func(self):
        "This performs the actual command."
        "Set the attacker to the caller and the defender to the target."

        if not self.rules.is_in_combat(self.caller):  # If not in combat, can't attack.
            self.caller.msg("You can only do that in combat. (see: help fight)")
            return

        if not self.rules.is_turn(self.caller):  # If it's not your turn, can't attack.
            self.caller.msg("You can only do that on your turn.")
            return

        if not self.caller.db.hp:  # Can't attack if you have no HP.
            self.caller.msg("You can't attack, you've been defeated.")
            return

        attacker = self.caller
        defender = self.caller.search(self.args)

        if not defender:  # No valid target given.
            return

        if not defender.db.hp:  # Target object has no HP left or to begin with
            self.caller.msg("You can't fight that!")
            return

        if attacker == defender:  # Target and attacker are the same
            self.caller.msg("You can't attack yourself!")
            return

        # Test to see if there are any nearby enemy targets.
        in_melee = []
        for target in attacker.db.combat_range:
            # Object is engaged and has HP
            if (
                self.rules.get_range(attacker, defender) == 0
                and target.db.hp
                and target != self.caller
            ):
                in_melee.append(target)  # Add to list of targets in melee

        if len(in_melee) > 0:
            self.caller.msg(
                "You can't shoot because there are fighters engaged with you (%s) - you need "
                "to retreat! (see: help withdraw)" % ", ".join(obj.key for obj in in_melee)
            )
            return

        "If everything checks out, call the attack resolving function."
        self.rules.resolve_attack(attacker, defender, "ranged")
        self.rules.spend_action(self.caller, 1, action_name="attack")  # Use up one action.


class CmdApproach(Command):
    """
    Approaches an object.

    Usage:
      approach <target>

    Move one space toward a character or object. You can only attack
    characters you are 0 spaces away from.
    """

    key = "approach"
    help_category = "combat"

    rules = COMBAT_RULES

    def func(self):
        "This performs the actual command."

        if not self.rules.is_in_combat(self.caller):  # If not in combat, can't approach.
            self.caller.msg("You can only do that in combat. (see: help fight)")
            return

        if not self.rules.is_turn(self.caller):  # If it's not your turn, can't approach.
            self.caller.msg("You can only do that on your turn.")
            return

        if not self.caller.db.hp:  # Can't approach if you have no HP.
            self.caller.msg("You can't move, you've been defeated.")
            return

        mover = self.caller
        target = self.caller.search(self.args)

        if not target:  # No valid target given.
            return

        if not target.db.combat_range:  # Target object is not on the range field
            self.caller.msg("You can't move toward that!")
            return

        if mover == target:  # Target and mover are the same
            self.caller.msg("You can't move toward yourself!")
            return

        if self.rules.get_range(mover, target) <= 0:  # Already engaged with target
            self.caller.msg("You're already next to that target!")
            return

        # If everything checks out, call the approach resolving function.
        self.rules.approach(mover, target)
        mover.location.msg_contents("%s moves toward %s." % (mover, target))
        self.rules.spend_action(self.caller, 1, action_name="move")  # Use up one action.


class CmdWithdraw(Command):
    """
    Moves away from an object.

    Usage:
      withdraw <target>

    Move one space away from a character or object.
    """

    key = "withdraw"
    help_category = "combat"

    rules = COMBAT_RULES

    def func(self):
        "This performs the actual command."

        if not self.rules.is_in_combat(self.caller):  # If not in combat, can't withdraw.
            self.caller.msg("You can only do that in combat. (see: help fight)")
            return

        if not self.rules.is_turn(self.caller):  # If it's not your turn, can't withdraw.
            self.caller.msg("You can only do that on your turn.")
            return

        if not self.caller.db.hp:  # Can't withdraw if you have no HP.
            self.caller.msg("You can't move, you've been defeated.")
            return

        mover = self.caller
        target = self.caller.search(self.args)

        if not target:  # No valid target given.
            return

        if not target.db.combat_range:  # Target object is not on the range field
            self.caller.msg("You can't move away from that!")
            return

        if mover == target:  # Target and mover are the same
            self.caller.msg("You can't move away from yourself!")
            return

        if mover.db.combat_range[target] >= 3:  # Already at maximum distance
            self.caller.msg("You're as far as you can get from that target!")
            return

        # If everything checks out, call the approach resolving function.
        self.rules.withdraw(mover, target)
        mover.location.msg_contents("%s moves away from %s." % (mover, target))
        self.rules.spend_action(self.caller, 1, action_name="move")  # Use up one action.


class CmdPass(tb_basic.CmdPass):
    """
    Passes on your turn.

    Usage:
      pass

    When in a fight, you can use this command to end your turn early, even
    if there are still any actions you can take.
    """

    key = "pass"
    aliases = ["wait", "hold"]
    help_category = "combat"

    rules = COMBAT_RULES


class CmdDisengage(tb_basic.CmdDisengage):
    """
    Passes your turn and attempts to end combat.

    Usage:
      disengage

    Ends your turn early and signals that you're trying to end
    the fight. If all participants in a fight disengage, the
    fight ends.
    """

    key = "disengage"
    aliases = ["spare"]
    help_category = "combat"

    rules = COMBAT_RULES


class CmdRest(tb_basic.CmdRest):
    """
    Recovers damage.

    Usage:
      rest

    Resting recovers your HP to its maximum, but you can only
    rest if you're not in a fight.
    """

    key = "rest"
    help_category = "combat"

    rules = COMBAT_RULES


class CmdStatus(Command):
    """
    Gives combat information.

    Usage:
      status

    Shows your current and maximum HP and your distance from
    other targets in combat.
    """

    key = "status"
    help_category = "combat"

    rules = COMBAT_RULES

    def func(self):
        "This performs the actual command."
        self.rules.combat_status_message(self.caller)


class CmdCombatHelp(tb_basic.CmdCombatHelp):
    """
    View help or a list of topics

    Usage:
      help <topic or command>
      help list
      help all

    This will search for help on commands and other
    topics related to the game.
    """

    # Just like the default help command, but will give quick
    # tips on combat when used in a fight with no arguments.
    rules = COMBAT_RULES
    combat_help_text = (
        "Available combat commands:|/"
        "|wAttack:|n Attack an engaged target, attempting to deal damage.|/"
        "|wShoot:|n Attack from a distance, if not engaged with other fighters.|/"
        "|wApproach:|n Move one step cloer to a target.|/"
        "|wWithdraw:|n Move one step away from a target.|/"
        "|wPass:|n Pass your turn without further action.|/"
        "|wStatus:|n View current HP and ranges to other targets.|/"
        "|wDisengage:|n End your turn and attempt to end combat.|/"
    )


class BattleCmdSet(default_cmds.CharacterCmdSet):
    """
    This command set includes all the commmands used in the battle system.
    """

    key = "DefaultCharacter"

    def at_cmdset_creation(self):
        """
        Populates the cmdset
        """
        self.add(CmdFight())
        self.add(CmdAttack())
        self.add(CmdShoot())
        self.add(CmdRest())
        self.add(CmdPass())
        self.add(CmdDisengage())
        self.add(CmdApproach())
        self.add(CmdWithdraw())
        self.add(CmdStatus())
        self.add(CmdCombatHelp())



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-06 17:31


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("help", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="helpentry",
            name="db_tags",
            field=models.ManyToManyField(
                blank=True,
                help_text="tags on this object. Tags are simple string markers to identify, group and alias objects.",
                to="typeclasses.Tag",
            ),
        )
    ]



"""
Service for integrating the Evennia Game Index client into Evennia.

"""
from twisted.application.service import Service
from twisted.internet import reactor
from twisted.internet.task import LoopingCall

from evennia.utils import logger

from .client import EvenniaGameIndexClient

# How many seconds to wait before triggering the first EGI check-in.
_FIRST_UPDATE_DELAY = 10
# How often to sync to the server
_CLIENT_UPDATE_RATE = 60 * 30


class EvenniaGameIndexService(Service):
    """
    Twisted Service that contains a LoopingCall for regularly sending game details
    to the Evennia Game Index.

    """

    # We didn't stick the Evennia prefix on here because it'd get marked as
    # a core system service.
    name = "GameIndexClient"

    def __init__(self):
        self.client = EvenniaGameIndexClient(on_bad_request=self._die_on_bad_request)
        self.loop = LoopingCall(self.client.send_game_details)

    def startService(self):
        super().startService()
        # Check to make sure that the client is configured.
        # Start the loop, but only after a short delay. This allows the
        # portal and the server time to sync up as far as total player counts.
        # Prevents always reporting a count of 0.
        reactor.callLater(_FIRST_UPDATE_DELAY, self.loop.start, _CLIENT_UPDATE_RATE)

    def stopService(self):
        if self.running == 0:
            # reload errors if we've stopped this service.
            return
        super().stopService()
        if self.loop.running:
            self.loop.stop()

    def _die_on_bad_request(self):
        """
        If it becomes apparent that our configuration is generating improperly
        formed messages to EGI, we don't want to keep sending bad messages.
        Stop the service so we're not wasting resources.
        """
        logger.log_infomsg(
            "Shutting down Evennia Game Index client service due to invalid configuration."
        )
        self.stopService()



"""
Custom manager for ServerConfig objects.
"""
from django.db import models


class ServerConfigManager(models.Manager):
    """
    This ServerConfigManager implements methods for searching and
    manipulating ServerConfigs directly from the database.

    These methods will all return database objects (or QuerySets)
    directly.

    ServerConfigs are used to store certain persistent settings for
    the server at run-time.

    """

    def conf(self, key=None, value=None, delete=False, default=None):
        """
        Add, retrieve and manipulate config values.

        Args:
            key (str, optional): Name of config.
            value (str, optional): Data to store in this config value.
            delete (bool, optional): If `True`, delete config with `key`.
            default (str, optional): Use when retrieving a config value
                by a key that does not exist.
        Returns:
            all (list): If `key` was not given - all stored config values.
            value (str): If `key` was given, this is the stored value, or
                `default` if no matching `key` was found.

        """
        if not key:
            return self.all()
        elif delete is True:
            for conf in self.filter(db_key=key):
                conf.delete()
        elif value is not None:
            conf = self.filter(db_key=key)
            if conf:
                conf = conf[0]
            else:
                conf = self.model(db_key=key)
            conf.value = value  # this will pickle
        else:
            conf = self.filter(db_key=key)
            if not conf:
                if callable(default):
                    # allows for `dict` to be a default value
                    return default()
                return default
            return conf[0].value
        return None



"""
This file defines global variables that will always be available in a view
context without having to repeatedly include it.

For this to work, this file is included in the settings file, in the
TEMPLATES["OPTIONS"]["context_processors"] list.

"""


import os

from django.conf import settings

from evennia.utils.utils import get_evennia_version

# Setup lists of the most relevant apps so
# the adminsite becomes more readable.

GAME_NAME = None
GAME_SLOGAN = None
SERVER_VERSION = None
SERVER_HOSTNAME = None

REGISTER_ENABLED = None

TELNET_ENABLED = None
TELNET_PORTS = None
TELNET_SSL_ENABLED = None
TELNET_SSL_PORTS = None

SSH_ENABLED = None
SSH_PORTS = None

WEBCLIENT_ENABLED = None
WEBSOCKET_CLIENT_ENABLED = None
WEBSOCKET_PORT = None
WEBSOCKET_URL = None

REST_API_ENABLED = False

ACCOUNT_RELATED = ["Accounts"]
GAME_ENTITIES = ["Objects", "Scripts", "Comms", "Help"]
GAME_SETUP = ["Permissions", "Config"]
CONNECTIONS = ["Irc"]
WEBSITE = ["Flatpages", "News", "Sites"]


def load_game_settings():
    """
    Load and cache game settings.

    """
    global GAME_NAME, GAME_SLOGAN, SERVER_VERSION, SERVER_HOSTNAME
    global REGISTER_ENABLED
    global TELNET_ENABLED, TELNET_PORTS
    global TELNET_SSL_ENABLED, TELNET_SSL_PORTS
    global SSH_ENABLED, SSH_PORTS
    global WEBCLIENT_ENABLED, WEBSOCKET_CLIENT_ENABLED, WEBSOCKET_PORT, WEBSOCKET_URL
    global REST_API_ENABLED

    try:
        GAME_NAME = settings.SERVERNAME.strip()
    except AttributeError:
        GAME_NAME = "Evennia"
    SERVER_VERSION = get_evennia_version()
    try:
        GAME_SLOGAN = settings.GAME_SLOGAN.strip()
    except AttributeError:
        GAME_SLOGAN = SERVER_VERSION
    SERVER_HOSTNAME = settings.SERVER_HOSTNAME

    REGISTER_ENABLED = settings.NEW_ACCOUNT_REGISTRATION_ENABLED

    TELNET_ENABLED = settings.TELNET_ENABLED
    TELNET_PORTS = settings.TELNET_PORTS
    TELNET_SSL_ENABLED = settings.SSL_ENABLED
    TELNET_SSL_PORTS = settings.SSL_PORTS

    SSH_ENABLED = settings.SSH_ENABLED
    SSH_PORTS = settings.SSH_PORTS

    WEBCLIENT_ENABLED = settings.WEBCLIENT_ENABLED
    WEBSOCKET_CLIENT_ENABLED = settings.WEBSOCKET_CLIENT_ENABLED
    # if we are working through a proxy or uses docker port-remapping, the webclient port encoded
    # in the webclient should be different than the one the server expects. Use the environment
    # variable WEBSOCKET_CLIENT_PROXY_PORT if this is the case.
    WEBSOCKET_PORT = int(
        os.environ.get("WEBSOCKET_CLIENT_PROXY_PORT", settings.WEBSOCKET_CLIENT_PORT)
    )
    # this is determined dynamically by the client and is less of an issue
    WEBSOCKET_URL = settings.WEBSOCKET_CLIENT_URL

    REST_API_ENABLED = settings.REST_API_ENABLED


load_game_settings()


# The main context processor function
def general_context(request):
    """
    Returns common Evennia-related context stuff, which is automatically added
    to context of all views.

    """
    account = None
    if request.user.is_authenticated:
        account = request.user

    puppet = None
    if account and request.session.get("puppet"):
        pk = int(request.session.get("puppet"))
        puppet = next((x for x in account.characters if x.pk == pk), None)

    return {
        "account": account,
        "puppet": puppet,
        "game_name": GAME_NAME,
        "game_slogan": GAME_SLOGAN,
        "server_hostname": SERVER_HOSTNAME,
        "evennia_userapps": ACCOUNT_RELATED,
        "evennia_entityapps": GAME_ENTITIES,
        "evennia_setupapps": GAME_SETUP,
        "evennia_connectapps": CONNECTIONS,
        "evennia_websiteapps": WEBSITE,
        "register_enabled": REGISTER_ENABLED,
        "telnet_enabled": TELNET_ENABLED,
        "telnet_ports": TELNET_PORTS,
        "telnet_ssl_enabled": TELNET_SSL_ENABLED,
        "telnet_ssl_ports": TELNET_SSL_PORTS,
        "ssh_enabled": SSH_ENABLED,
        "ssh_ports": SSH_ENABLED,
        "webclient_enabled": WEBCLIENT_ENABLED,
        "websocket_enabled": WEBSOCKET_CLIENT_ENABLED,
        "websocket_port": WEBSOCKET_PORT,
        "websocket_url": WEBSOCKET_URL,
        "rest_api_enabled": REST_API_ENABLED,
    }



"""
Sets up an api-access permission check using the in-game permission hierarchy.

"""


from django.conf import settings
from rest_framework import permissions

from evennia.locks.lockhandler import check_perm


class EvenniaPermission(permissions.BasePermission):
    """
    A Django Rest Framework permission class that allows us to use Evennia's
    permission structure. Based on the action in a given view, we'll check a
    corresponding Evennia access/lock check.

    """

    # subclass this to change these permissions
    MINIMUM_LIST_PERMISSION = settings.REST_FRAMEWORK.get("DEFAULT_LIST_PERMISSION", "builder")
    MINIMUM_CREATE_PERMISSION = settings.REST_FRAMEWORK.get("DEFAULT_CREATE_PERMISSION", "builder")
    view_locks = settings.REST_FRAMEWORK.get("DEFAULT_VIEW_LOCKS", ["examine"])
    destroy_locks = settings.REST_FRAMEWORK.get("DEFAULT_DESTROY_LOCKS", ["delete"])
    update_locks = settings.REST_FRAMEWORK.get("DEFAULT_UPDATE_LOCKS", ["control", "edit"])

    def has_permission(self, request, view):
        """Checks for permissions

        Args:
            request (Request): The incoming request object.
            view (View): The django view we are checking permission for.

        Returns:
            bool: If permission is granted or not. If we return False here, a PermissionDenied
            error will be raised from the view.

        Notes:
            This method is a check that always happens first. If there's an object involved,
            such as with retrieve, update, or delete, then the has_object_permission method
            is called after this, assuming this returns `True`.
        """
        # Only allow authenticated users to call the API
        if not request.user.is_authenticated:
            return False
        if request.user.is_superuser:
            return True
        # these actions don't support object-level permissions, so use the above definitions
        if view.action == "list":
            return check_perm(request.user, self.MINIMUM_LIST_PERMISSION)
        if view.action == "create":
            return check_perm(request.user, self.MINIMUM_CREATE_PERMISSION)
        return True  # this means we'll check object-level permissions

    @staticmethod
    def check_locks(obj, user, locks):
        """Checks access for user for object with given locks
        Args:
            obj: Object instance we're checking
            user (Account): User who we're checking permissions
            locks (list): list of lockstrings

        Returns:
            bool: True if they have access, False if they don't
        """
        return any([obj.access(user, lock) for lock in locks])

    def has_object_permission(self, request, view, obj):
        """Checks object-level permissions after has_permission

        Args:
            request (Request): The incoming request object.
            view (View): The django view we are checking permission for.
            obj: Object we're checking object-level permissions for

        Returns:
            bool: If permission is granted or not. If we return False here, a PermissionDenied
            error will be raised from the view.

        Notes:
            This method assumes that has_permission has already returned True. We check
            equivalent Evennia permissions in the request.user to determine if they can
            complete the action.
        """
        if view.action in ("list", "retrieve"):
            # access_type is based on the examine command
            return self.check_locks(obj, request.user, self.view_locks)
        if view.action == "destroy":
            # access type based on the destroy command
            return self.check_locks(obj, request.user, self.destroy_locks)
        if view.action in ("update", "partial_update", "set_attribute"):
            # access type based on set command
            return self.check_locks(obj, request.user, self.update_locks)



from django.utils.translation import gettext as _

from evennia.utils.containers import OPTION_CLASSES
from evennia.utils.utils import string_partial_matching

_GA = object.__getattribute__
_SA = object.__setattr__


class InMemorySaveHandler:
    """
    Fallback SaveHandler, implementing a minimum of the required save mechanism
    and storing data in memory.

    """

    def __init__(self):
        self.storage = {}

    def add(self, key, value=None, **kwargs):
        self.storage[key] = value

    def get(self, key, default=None, **kwargs):
        return self.storage.get(key, default)


class OptionHandler:
    """
    This is a generic Option handler.  Retrieve options either as properties on
    this handler or by using the .get method.

    This is used for Account.options but it could be used by Scripts or Objects
    just as easily. All it needs to be provided is an options_dict.

    """

    def __init__(
        self,
        obj,
        options_dict=None,
        savefunc=None,
        loadfunc=None,
        save_kwargs=None,
        load_kwargs=None,
    ):
        """
        Initialize an OptionHandler.

        Args:
            obj (object): The object this handler sits on. This is usually a TypedObject.
            options_dict (dict): A dictionary of option keys, where the values
                are options. The format of those tuples is: ('key', "Description to
                show", 'option_type', <default value>)
            savefunc (callable): A callable for all options to call when saving itself.
                It will be called as `savefunc(key, value, **save_kwargs)`. A common one
                to pass would be AttributeHandler.add.
            loadfunc (callable): A callable for all options to call when loading data into
                itself. It will be called as `loadfunc(key, default=default, **load_kwargs)`.
                A common one to pass would be AttributeHandler.get.
            save_kwargs (any): Optional extra kwargs to pass into `savefunc` above.
            load_kwargs (any): Optional extra kwargs to pass into `loadfunc` above.

        Notes:
            Both loadfunc and savefunc must be specified. If only one is given, the other
            will be ignored and in-memory storage will be used.

        """
        self.obj = obj
        self.options_dict = {} if options_dict is None else options_dict

        if not savefunc and loadfunc:
            self._in_memory_handler = InMemorySaveHandler()
            savefunc = InMemorySaveHandler.add
            loadfunc = InMemorySaveHandler.get
        self.savefunc = savefunc
        self.loadfunc = loadfunc
        self.save_kwargs = {} if save_kwargs is None else save_kwargs
        self.load_kwargs = {} if load_kwargs is None else load_kwargs

        # This dictionary stores the in-memory Options objects by their key for
        # quick lookup.
        self.options = {}

    def __getattr__(self, key):
        """
        Allow for obj.options.key

        """
        return self.get(key)

    def __setattr__(self, key, value):
        """
        Allow for obj.options.key = value

        But we must be careful to avoid infinite loops!

        """
        try:
            if key in _GA(self, "options_dict"):
                _GA(self, "set")(key, value)
        except AttributeError:
            pass
        _SA(self, key, value)

    def _load_option(self, key):
        """
        Loads option on-demand if it has not been loaded yet.

        Args:
            key (str): The option being loaded.

        Returns:

        """
        desc, clsname, default_val = self.options_dict[key]
        loaded_option = OPTION_CLASSES.get(clsname)(self, key, desc, default_val)
        # store the value for future easy access
        self.options[key] = loaded_option
        return loaded_option

    def get(self, key, default=None, return_obj=False, raise_error=False):
        """
        Retrieves an Option stored in the handler. Will load it if it doesn't exist.

        Args:
            key (str): The option key to retrieve.
            default (any): What to return if the option is defined.
            return_obj (bool, optional): If True, returns the actual option
                object instead of its value.
            raise_error (bool, optional): Raise Exception if key is not found in options.
        Returns:
            option_value (any or Option): An option value  the Option itself.
        Raises:
            KeyError: If option is not defined.

        """
        if key not in self.options_dict:
            if raise_error:
                raise KeyError(_("Option not found!"))
            return default
        # get the options or load/recache it
        op_found = self.options.get(key) or self._load_option(key)
        return op_found if return_obj else op_found.value

    def set(self, key, value, **kwargs) -> "BaseOption":
        """
        Change an individual option.

        Args:
            key (str): The key of an option that can be changed. Allows partial matching.
            value (str): The value that should be checked, coerced, and stored.:
            kwargs (any, optional): These are passed into the Option's validation function,
                save function and display function and allows to customize either.

        Returns:
            BaseOption: The matched object. Its new value can be accessed with
                op.value or op.display().

        """
        if not key:
            raise ValueError(_("Option field blank!"))
        match = string_partial_matching(list(self.options_dict.keys()), key, ret_index=False)
        if not match:
            raise ValueError(_("Option not found!"))
        if len(match) > 1:
            raise ValueError(
                _("Multiple matches:") + f"{', '.join(match)}. " + _("Please be more specific.")
            )
        match = match[0]
        op = self.get(match, return_obj=True)
        op.set(value, **kwargs)
        return op

    def all(self, return_objs=False):
        """
        Get all options defined on this handler.

        Args:
            return_objs (bool, optional): Return the actual Option objects rather
                than their values.
        Returns:
            all_options (dict): All options on this handler, either `{key: value}`
                or `{key: <Option>}` if `return_objs` is `True`.

        """
        return [self.get(key, return_obj=return_objs) for key in self.options_dict]



"""
Auditable Server Sessions:
Extension of the stock ServerSession that yields objects representing
user inputs and system outputs.

Evennia contribution - Johnny 2017
"""
import os
import re
import socket

from django.conf import settings as ev_settings
from django.utils import timezone

from evennia.server.serversession import ServerSession
from evennia.utils import get_evennia_version, logger, mod_import, utils

# Attributes governing auditing of commands and where to send log objects
AUDIT_CALLBACK = getattr(
    ev_settings, "AUDIT_CALLBACK", "evennia.contrib.utils.auditing.outputs.to_file"
)
AUDIT_IN = getattr(ev_settings, "AUDIT_IN", False)
AUDIT_OUT = getattr(ev_settings, "AUDIT_OUT", False)
AUDIT_ALLOW_SPARSE = getattr(ev_settings, "AUDIT_ALLOW_SPARSE", False)
AUDIT_MASKS = [
    {"connect": r"^[@\s]*[connect]{5,8}\s+(\".+?\"|[^\s]+)\s+(?P<secret>.+)"},
    {"connect": r"^[@\s]*[connect]{5,8}\s+(?P<secret>[\w]+)"},
    {"create": r"^[^@]?[create]{5,6}\s+(\w+|\".+?\")\s+(?P<secret>[\w]+)"},
    {"create": r"^[^@]?[create]{5,6}\s+(?P<secret>[\w]+)"},
    {"userpassword": r"^[@\s]*[userpassword]{11,14}\s+(\w+|\".+?\")\s+=*\s*(?P<secret>[\w]+)"},
    {"userpassword": r"^.*new password set to '(?P<secret>[^']+)'\."},
    {"userpassword": r"^.* has changed your password to '(?P<secret>[^']+)'\."},
    {"password": r"^[@\s]*[password]{6,9}\s+(?P<secret>.*)"},
] + getattr(ev_settings, "AUDIT_MASKS", [])


if AUDIT_CALLBACK:
    try:
        AUDIT_CALLBACK = getattr(
            mod_import(".".join(AUDIT_CALLBACK.split(".")[:-1])), AUDIT_CALLBACK.split(".")[-1]
        )
        logger.log_sec("Auditing module online.")
        logger.log_sec(
            "Audit record User input: {}, output: {}.\n"
            "Audit sparse recording: {}, Log callback: {}".format(
                AUDIT_IN, AUDIT_OUT, AUDIT_ALLOW_SPARSE, AUDIT_CALLBACK
            )
        )
    except Exception as e:
        logger.log_err("Failed to activate Auditing module. %s" % e)


class AuditedServerSession(ServerSession):
    """
    This particular implementation parses all server inputs and/or outputs and
    passes a dict containing the parsed metadata to a callback method of your
    creation. This is useful for recording player activity where necessary for
    security auditing, usage analysis or post-incident forensic discovery.

    *** WARNING ***
    All strings are recorded and stored in plaintext. This includes those strings
    which might contain sensitive data (create, connect, @password). These commands
    have their arguments masked by default, but you must mask or mask any
    custom commands of your own that handle sensitive information.

    See README.md for installation/configuration instructions.
    """

    def audit(self, **kwargs):
        """
        Extracts messages and system data from a Session object upon message
        send or receive.

        Keyword Args:
            src (str): Source of data; 'client' or 'server'. Indicates direction.
            text (str or list): Client sends messages to server in the form of
                lists. Server sends messages to client as string.

        Returns:
            log (dict): Dictionary object containing parsed system and user data
                related to this message.

        """
        # Get time at start of processing
        time_obj = timezone.now()
        time_str = str(time_obj)

        session = self
        src = kwargs.pop("src", "?")
        bytecount = 0

        # Do not log empty lines
        if not kwargs:
            return {}

        # Get current session's IP address
        client_ip = session.address

        # Capture Account name and dbref together
        account = session.get_account()
        account_token = ""
        if account:
            account_token = "%s%s" % (account.key, account.dbref)

        # Capture Character name and dbref together
        char = session.get_puppet()
        char_token = ""
        if char:
            char_token = "%s%s" % (char.key, char.dbref)

        # Capture Room name and dbref together
        room = None
        room_token = ""
        if char:
            room = char.location
            room_token = "%s%s" % (room.key, room.dbref)

        # Try to compile an input/output string
        def drill(obj, bucket):
            if isinstance(obj, dict):
                return bucket
            elif utils.is_iter(obj):
                for sub_obj in obj:
                    bucket.extend(drill(sub_obj, []))
            else:
                bucket.append(obj)
            return bucket

        text = kwargs.pop("text", "")
        if utils.is_iter(text):
            text = "|".join(drill(text, []))

        # Mask any PII in message, where possible
        bytecount = len(text.encode("utf-8"))
        text = self.mask(text)

        # Compile the IP, Account, Character, Room, and the message.
        log = {
            "time": time_str,
            "hostname": socket.getfqdn(),
            "application": "%s" % ev_settings.SERVERNAME,
            "version": get_evennia_version(),
            "pid": os.getpid(),
            "direction": "SND" if src == "server" else "RCV",
            "protocol": self.protocol_key,
            "ip": client_ip,
            "session": "session#%s" % self.sessid,
            "account": account_token,
            "character": char_token,
            "room": room_token,
            "text": text.strip(),
            "bytes": bytecount,
            "data": kwargs,
            "objects": {
                "time": time_obj,
                "session": self,
                "account": account,
                "character": char,
                "room": room,
            },
        }

        # Remove any keys with blank values
        if AUDIT_ALLOW_SPARSE is False:
            log["data"] = {k: v for k, v in log["data"].items() if v}
            log["objects"] = {k: v for k, v in log["objects"].items() if v}
            log = {k: v for k, v in log.items() if v}

        return log

    def mask(self, msg):
        """
        Masks potentially sensitive user information within messages before
        writing to log. Recording cleartext password attempts is bad policy.

        Args:
            msg (str): Raw text string sent from client <-> server

        Returns:
            msg (str): Text string with sensitive information masked out.

        """
        # Check to see if the command is embedded within server output
        _msg = msg
        is_embedded = False
        match = re.match(".*Command.*'(.+)'.*is not available.*", msg, flags=re.IGNORECASE)
        if match:
            msg = match.group(1).replace("\\", "")
            submsg = msg
            is_embedded = True

        for mask in AUDIT_MASKS:
            for command, regex in mask.items():
                try:
                    match = re.match(regex, msg, flags=re.IGNORECASE)
                except Exception as e:
                    logger.log_err(regex)
                    logger.log_err(e)
                    continue

                if match:
                    term = match.group("secret")
                    masked = re.sub(term, "*" * len(term.zfill(8)), msg)

                    if is_embedded:
                        msg = re.sub(
                            submsg, "%s <Masked: %s>" % (masked, command), _msg, flags=re.IGNORECASE
                        )
                    else:
                        msg = masked

                    return msg

        return _msg

    def data_out(self, **kwargs):
        """
        Generic hook for sending data out through the protocol.

        Keyword Args:
            kwargs (any): Other data to the protocol.

        """
        if AUDIT_CALLBACK and AUDIT_OUT:
            try:
                log = self.audit(src="server", **kwargs)
                if log:
                    AUDIT_CALLBACK(log)
            except Exception as e:
                logger.log_err(e)

        super().data_out(**kwargs)

    def data_in(self, **kwargs):
        """
        Hook for protocols to send incoming data to the engine.

        Keyword Args:
            kwargs (any): Other data from the protocol.

        """
        if AUDIT_CALLBACK and AUDIT_IN:
            try:
                log = self.audit(src="client", **kwargs)
                if log:
                    AUDIT_CALLBACK(log)
            except Exception as e:
                logger.log_err(e)

        super().data_in(**kwargs)



"""
This redirects to website sub-pages.

"""
from django.conf import settings
from django.contrib import admin
from django.urls import include, path

from .views import accounts, channels, characters, errors
from .views import help as helpviews
from .views import index

urlpatterns = [
    # website front page
    path("", index.EvenniaIndexView.as_view(), name="index"),
    # errors
    path(r"tbi/", errors.to_be_implemented, name="to_be_implemented"),
    # User Authentication (makes login/logout url names available)
    path("auth/register", accounts.AccountCreateView.as_view(), name="register"),
    path("auth/", include("django.contrib.auth.urls")),
    # Help Topics
    path("help/", helpviews.HelpListView.as_view(), name="help"),
    path(
        r"help/<str:category>/<str:topic>/",
        helpviews.HelpDetailView.as_view(),
        name="help-entry-detail",
    ),
    # Channels
    path("channels/", channels.ChannelListView.as_view(), name="channels"),
    path("channels/<str:slug>/", channels.ChannelDetailView.as_view(), name="channel-detail"),
    # Character management
    path("characters/", characters.CharacterListView.as_view(), name="characters"),
    path("characters/create/", characters.CharacterCreateView.as_view(), name="character-create"),
    path("characters/manage/", characters.CharacterManageView.as_view(), name="character-manage"),
    path(
        "characters/detail/<str:slug>/<int:pk>/",
        characters.CharacterDetailView.as_view(),
        name="character-detail",
    ),
    path(
        "characters/puppet/<str:slug>/<int:pk>/",
        characters.CharacterPuppetView.as_view(),
        name="character-puppet",
    ),
    path(
        "characters/update/<str:slug>/<int:pk>/",
        characters.CharacterUpdateView.as_view(),
        name="character-update",
    ),
    path(
        "characters/delete/<str:slug>/<int:pk>/",
        characters.CharacterDeleteView.as_view(),
        name="character-delete",
    ),
]


# This sets up the server if the user want to run the Django test server (this
# is not recommended and is usually unnecessary).
if settings.SERVE_MEDIA:
    from django import views as django_views

    urlpatterns.extend(
        [
            path(
                "media/<str:path>",
                django_views.static.serve,
                {"document_root": settings.MEDIA_ROOT},
            ),
            path(
                "static/<str:path>",
                django_views.static.serve,
                {"document_root": settings.STATIC_ROOT},
            ),
        ]
    )



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ServerConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID", serialize=False, auto_created=True, primary_key=True
                    ),
                ),
                ("db_key", models.CharField(unique=True, max_length=64)),
                ("db_value", models.TextField(blank=True)),
            ],
            options={
                "verbose_name": "Server Config value",
                "verbose_name_plural": "Server Config values",
            },
            bases=(models.Model,),
        )
    ]



"""
This module defines a generic session class. All connection instances
(both on Portal and Server side) should inherit from this class.

"""
import time

from django.conf import settings

# ------------------------------------------------------------
# Server Session
# ------------------------------------------------------------


class Session:
    """
    This class represents a player's session and is a template for
    both portal- and server-side sessions.

    Each connection will see two session instances created:

     1. A Portal session. This is customized for the respective connection
        protocols that Evennia supports, like Telnet, SSH etc. The Portal
        session must call init_session() as part of its initialization. The
        respective hook methods should be connected to the methods unique
        for the respective protocol so that there is a unified interface
        to Evennia.
     2. A Server session. This is the same for all connected accounts,
        regardless of how they connect.

    The Portal and Server have their own respective sessionhandlers. These
    are synced whenever new connections happen or the Server restarts etc,
    which means much of the same information must be stored in both places
    e.g. the portal can re-sync with the server when the server reboots.

    """

    def init_session(self, protocol_key, address, sessionhandler):
        """
        Initialize the Session. This should be called by the protocol when
        a new session is established.

        Args:
            protocol_key (str): By default, one of 'telnet', 'telnet/ssl', 'ssh',
                'webclient/websocket' or 'webclient/ajax'.
            address (str): Client address.
            sessionhandler (SessionHandler): Reference to the
                main sessionhandler instance.

        """
        # This is currently 'telnet', 'ssh', 'ssl' or 'web'
        self.protocol_key = protocol_key
        # Protocol address tied to this session
        self.address = address

        # suid is used by some protocols, it's a hex key.
        self.suid = None

        # unique id for this session
        self.sessid = 0  # no sessid yet
        # client session id, if given by the client
        self.csessid = None
        # database id for the user connected to this session
        self.uid = None
        # user name, for easier tracking of sessions
        self.uname = None
        # if user has authenticated already or not
        self.logged_in = False

        # database id of puppeted object (if any)
        self.puid = None

        # session time statistics
        self.conn_time = time.time()
        self.cmd_last_visible = self.conn_time
        self.cmd_last = self.conn_time
        self.cmd_total = 0

        self.protocol_flags = {
            "ENCODING": "utf-8",
            "SCREENREADER": False,
            "INPUTDEBUG": False,
            "RAW": False,
            "NOCOLOR": False,
            "LOCALECHO": False,
        }
        self.server_data = {}

        # map of input data to session methods
        self.datamap = {}

        # a back-reference to the relevant sessionhandler this
        # session is stored in.
        self.sessionhandler = sessionhandler

    def get_sync_data(self):
        """
        Get all data relevant to sync the session.

        Args:
            syncdata (dict): All syncdata values, based on
                the keys given by self._attrs_to_sync.

        """
        return {
            attr: getattr(self, attr) for attr in settings.SESSION_SYNC_ATTRS if hasattr(self, attr)
        }

    def load_sync_data(self, sessdata):
        """
        Takes a session dictionary, as created by get_sync_data, and
        loads it into the correct properties of the session.

        Args:
            sessdata (dict): Session data dictionary.

        """
        for propname, value in sessdata.items():
            if (
                propname == "protocol_flags"
                and isinstance(value, dict)
                and hasattr(self, "protocol_flags")
                and isinstance(self.protocol_flags, dict)
            ):
                # special handling to allow partial update of protocol flags
                self.protocol_flags.update(value)
            else:
                setattr(self, propname, value)

    def at_sync(self):
        """
        Called after a session has been fully synced (including
        secondary operations such as setting self.account based
        on uid etc).

        """
        if self.account:
            self.protocol_flags.update(
                self.account.attributes.get("_saved_protocol_flags", None) or {}
            )

    # access hooks

    def disconnect(self, reason=None):
        """
        generic hook called from the outside to disconnect this session
        should be connected to the protocols actual disconnect mechanism.

        Args:
            reason (str): Eventual text motivating the disconnect.

        """
        pass

    def data_out(self, **kwargs):
        """
        Generic hook for sending data out through the protocol. Server
        protocols can use this right away. Portal sessions
        should overload this to format/handle the outgoing data as needed.

        Keyword Args:
            kwargs (any): Other data to the protocol.

        """
        pass

    def data_in(self, **kwargs):
        """
        Hook for protocols to send incoming data to the engine.

        Keyword Args:
            kwargs (any): Other data from the protocol.

        """
        pass



"""
Tag admin

"""


import traceback
from datetime import datetime

from django import forms
from django.contrib import admin

from evennia.typeclasses.tags import Tag
from evennia.utils.dbserialize import _SaverSet, from_pickle
from evennia.utils.picklefield import PickledFormField


class TagForm(forms.ModelForm):
    """
    Form to display fields in the stand-alone Tag display.

    """

    db_key = forms.CharField(label="Key/Name", required=True, help_text="The main key identifier")
    db_category = forms.CharField(
        label="Category",
        help_text="Used for grouping tags. Unset (default) gives a category of None",
        required=False,
    )
    db_tagtype = forms.ChoiceField(
        label="Type",
        choices=[(None, "-"), ("alias", "alias"), ("permission", "permission")],
        help_text="Tags are used for different things. Unset for regular tags.",
        required=False,
    )
    db_model = forms.ChoiceField(
        label="Model",
        required=False,
        help_text="Each Tag can only 'attach' to one type of entity.",
        choices=(
            [
                ("objectdb", "objectdb"),
                ("accountdb", "accountdb"),
                ("scriptdb", "scriptdb"),
                ("channeldb", "channeldb"),
                ("helpentry", "helpentry"),
                ("msg", "msg"),
            ]
        ),
    )
    db_data = forms.CharField(
        label="Data",
        help_text="Usually unused. Intended for info about the tag itself",
        widget=forms.Textarea(attrs={"cols": "100", "rows": "2"}),
        required=False,
    )

    class Meta:
        fields = ("tag_key", "tag_category", "tag_data", "tag_type")


class InlineTagForm(forms.ModelForm):
    """
    Form for displaying tags inline together with other entities.

    This form overrides the base behavior of the ModelForm that would be used for a
    Tag-through-model.  Since the through-models only have access to the foreignkeys of the Tag and
    the Object that they're attached to, we need to spoof the behavior of it being a form that would
    correspond to its tag, or the creation of a tag. Instead of being saved, we'll call to the
    Object's handler, which will handle the creation, change, or deletion of a tag for us, as well
    as updating the handler's cache so that all changes are instantly updated in-game.
    """

    tag_key = forms.CharField(
        label="Tag Name", required=True, help_text="This is the main key identifier"
    )
    tag_category = forms.CharField(
        label="Category",
        help_text="Used for grouping tags. Unset (default) gives a category of None",
        required=False,
    )
    tag_type = forms.ChoiceField(
        label="Type",
        choices=[(None, "-"), ("alias", "alias"), ("permission", "permission")],
        help_text="Tags are used for different things. Unset for regular tags.",
        required=False,
    )
    tag_data = forms.CharField(
        label="Data",
        widget=forms.Textarea(attrs={"cols": "100", "rows": "2"}),
        help_text="Usually unused. Intended for eventual info about the tag itself",
        required=False,
    )

    class Meta:
        fields = ("tag_key", "tag_category", "tag_data", "tag_type")

    def __init__(self, *args, **kwargs):
        """
        If we have a tag, then we'll prepopulate our instance with the fields we'd expect it
        to have based on the tag. tag_key, tag_category, tag_type, and tag_data all refer to
        the corresponding tag fields. The initial data of the form fields will similarly be
        populated.
        """
        super().__init__(*args, **kwargs)
        tagkey = None
        tagcategory = None
        tagtype = None
        tagdata = None
        if hasattr(self.instance, "tag"):
            tagkey = self.instance.tag.db_key
            tagcategory = self.instance.tag.db_category
            tagtype = self.instance.tag.db_tagtype
            tagdata = self.instance.tag.db_data
            self.fields["tag_key"].initial = tagkey
            self.fields["tag_category"].initial = tagcategory
            self.fields["tag_type"].initial = tagtype
            self.fields["tag_data"].initial = tagdata
        self.instance.tag_key = tagkey
        self.instance.tag_category = tagcategory
        self.instance.tag_type = tagtype
        self.instance.tag_data = tagdata

    def save(self, commit=True):
        """
        One thing we want to do here is the or None checks, because forms are saved with an empty
        string rather than null from forms, usually, and the Handlers may handle empty strings
        differently than None objects. So for consistency with how things are handled in game,
        we'll try to make sure that empty form fields will be None, rather than ''.
        """
        # we are spoofing a tag for the Handler that will be called
        # instance = super().save(commit=False)
        instance = self.instance
        instance.tag_key = self.cleaned_data["tag_key"]
        instance.tag_category = self.cleaned_data["tag_category"] or None
        instance.tag_type = self.cleaned_data["tag_type"] or None
        instance.tag_data = self.cleaned_data["tag_data"] or None
        return instance


class TagFormSet(forms.BaseInlineFormSet):
    """
    The Formset handles all the inline forms that are grouped together on the change page of the
    corresponding object. All the tags will appear here, and we'll save them by overriding the
    formset's save method. The forms will similarly spoof their save methods to return an instance
    which hasn't been saved to the database, but have the relevant fields filled out based on the
    contents of the cleaned form. We'll then use that to call to the handler of the corresponding
    Object, where the handler is an AliasHandler, PermissionsHandler, or TagHandler, based on the
    type of tag.
    """

    verbose_name = "Tag"
    verbose_name_plural = "Tags"

    def save(self, commit=True):
        def get_handler(finished_object):
            related = getattr(finished_object, self.related_field)
            try:
                tagtype = finished_object.tag_type
            except AttributeError:
                tagtype = finished_object.tag.db_tagtype
            if tagtype == "alias":
                handler_name = "aliases"
            elif tagtype == "permission":
                handler_name = "permissions"
            else:
                handler_name = "tags"
            return getattr(related, handler_name)

        instances = super().save(commit=False)
        # self.deleted_objects is a list created when super of save is called, we'll remove those
        for obj in self.deleted_objects:
            handler = get_handler(obj)
            handler.remove(obj.tag_key, category=obj.tag_category)
        for instance in instances:
            handler = get_handler(instance)
            handler.add(instance.tag_key, category=instance.tag_category, data=instance.tag_data)


class TagInline(admin.TabularInline):
    """
    A handler for inline Tags. This class should be subclassed in the admin of your models,
    and the 'model' and 'related_field' class attributes must be set. model should be the
    through model (ObjectDB_db_tag', for example), while related field should be the name
    of the field on that through model which points to the model being used: 'objectdb',
    'msg', 'accountdb', etc.
    """

    # Set this to the through model of your desired M2M when subclassing.
    model = None
    verbose_name = "Tag"
    verbose_name_plural = "Tags"
    form = InlineTagForm
    formset = TagFormSet
    related_field = None  # Must be 'objectdb', 'accountdb', 'msg', etc. Set when subclassing
    # raw_id_fields = ('tag',)
    # readonly_fields = ('tag',)
    extra = 0

    def get_formset(self, request, obj=None, **kwargs):
        """
        get_formset has to return a class, but we need to make the class that we return
        know about the related_field that we'll use. Returning the class itself rather than
        a proxy isn't threadsafe, since it'd be the base class and would change if multiple
        people used the admin at the same time
        """
        formset = super().get_formset(request, obj, **kwargs)

        class ProxyFormset(formset):
            pass

        ProxyFormset.related_field = self.related_field
        return ProxyFormset


@admin.register(Tag)
class TagAdmin(admin.ModelAdmin):
    """
    A django Admin wrapper for Tags.

    """

    search_fields = ("db_key", "db_category", "db_tagtype")
    list_display = ("db_key", "db_category", "db_tagtype", "db_model", "db_data")
    list_filter = ("db_tagtype", "db_category", "db_model")
    form = TagForm
    view_on_site = False

    fieldsets = (
        (
            None,
            {"fields": (("db_key", "db_category"), ("db_tagtype", "db_model"), "db_data")},
        ),
    )



"""
Contains all the validation functions.

All validation functions must have a checker (probably a session) and entry arg.

They can employ more paramters at your leisure.


"""

import datetime as _dt
import re as _re

import pytz as _pytz
from django.utils.translation import gettext as _

from evennia.utils.ansi import strip_ansi
from evennia.utils.utils import string_partial_matching as _partial
from evennia.utils.utils import validate_email_address

_TZ_DICT = {str(tz): _pytz.timezone(tz) for tz in _pytz.common_timezones}


def text(entry, option_key="Text", **kwargs):
    try:
        return str(entry)
    except Exception as err:
        raise ValueError(_("Input could not be converted to text ({err})").format(err=err))


def color(entry, option_key="Color", **kwargs):
    """
    The color should be just a color character, so 'r' if red color is desired.

    """
    if not entry:
        raise ValueError(_("Nothing entered for a {option_key}!").format(option_key=option_key))
    test_str = strip_ansi(f"|{entry}|n")
    if test_str:
        raise ValueError(
            _("'{entry}' is not a valid {option_key}.").format(entry=entry, option_key=option_key)
        )
    return entry


def datetime(entry, option_key="Datetime", account=None, from_tz=None, **kwargs):
    """
    Process a datetime string in standard forms while accounting for the
    inputer's timezone. Always returns a result in UTC.

    Args:
        entry (str): A date string from a user.
        option_key (str): Name to display this datetime as.
        account (AccountDB): The Account performing this lookup. Unless `from_tz` is provided,
            the account's timezone option will be used.
        from_tz (pytz.timezone): An instance of a pytz timezone object from the
            user. If not provided, tries to use the timezone option of `account`.
            If neither one is provided, defaults to UTC.
    Returns:
        datetime in UTC.
    Raises:
        ValueError: If encountering a malformed timezone, date string or other format error.

    """
    if not entry:
        raise ValueError(_("No {option_key} entered!").format(option_key=option_key))
    if not from_tz:
        from_tz = _pytz.UTC
        if account:
            acct_tz = account.options.get("timezone", "UTC")
            try:
                from_tz = _pytz.timezone(acct_tz)
            except Exception as err:
                raise ValueError(
                    _("Timezone string '{acct_tz}' is not a valid timezone ({err})").format(
                        acct_tz=acct_tz, err=err
                    )
                )
        else:
            from_tz = _pytz.UTC

    utc = _pytz.UTC
    now = _dt.datetime.utcnow().replace(tzinfo=utc)
    cur_year = now.strftime("%Y")
    split_time = entry.split(" ")
    if len(split_time) == 3:
        entry = f"{split_time[0]} {split_time[1]} {split_time[2]} {cur_year}"
    elif len(split_time) == 4:
        entry = f"{split_time[0]} {split_time[1]} {split_time[2]} {split_time[3]}"
    else:
        raise ValueError(
            _("{option_key} must be entered in a 24-hour format such as: {timeformat}").format(
                option_key=option_key, timeformat=now.strftime("%b %d %H:%M")
            )
        )
    try:
        local = _dt.datetime.strptime(entry, "%b %d %H:%M %Y")
    except ValueError:
        raise ValueError(
            _("{option_key} must be entered in a 24-hour format such as: {timeformat}").format(
                option_key=option_key, timeformat=now.strftime("%b %d %H:%M")
            )
        )
    local_tz = from_tz.localize(local)
    return local_tz.astimezone(utc)


def duration(entry, option_key="Duration", **kwargs):
    """
    Take a string and derive a datetime timedelta from it.

    Args:
        entry (string): This is a string from user-input. The intended format is, for example:
            "5d 2w 90s" for 'five days, two weeks, and ninety seconds.' Invalid sections are
            ignored.
        option_key (str): Name to display this query as.

    Returns:
        timedelta

    """
    time_string = entry.lower().split(" ")
    seconds = 0
    minutes = 0
    hours = 0
    days = 0
    weeks = 0

    for interval in time_string:
        if _re.match(r"^[\d]+s$", interval):
            seconds += int(interval.rstrip("s"))
        elif _re.match(r"^[\d]+m$", interval):
            minutes += int(interval.rstrip("m"))
        elif _re.match(r"^[\d]+h$", interval):
            hours += int(interval.rstrip("h"))
        elif _re.match(r"^[\d]+d$", interval):
            days += int(interval.rstrip("d"))
        elif _re.match(r"^[\d]+w$", interval):
            weeks += int(interval.rstrip("w"))
        elif _re.match(r"^[\d]+y$", interval):
            days += int(interval.rstrip("y")) * 365
        else:
            raise ValueError(
                _("Could not convert section '{interval}' to a {option_key}.").format(
                    interval=interval, option_key=option_key
                )
            )

    return _dt.timedelta(days, seconds, 0, 0, minutes, hours, weeks)


def future(entry, option_key="Future Datetime", from_tz=None, **kwargs):
    time = datetime(entry, option_key, from_tz=from_tz)
    if time < _dt.datetime.utcnow().replace(tzinfo=_dt.timezone.utc):
        raise ValueError(
            _("That {option_key} is in the past! Must give a Future datetime!").format(
                option_key=option_key
            )
        )
    return time


def signed_integer(entry, option_key="Signed Integer", **kwargs):
    if not entry:
        raise ValueError(
            _("Must enter a whole number for {option_key}!").format(option_key=option_key)
        )
    try:
        num = int(entry)
    except ValueError:
        raise ValueError(
            _("Could not convert '{entry}' to a whole " "number for {option_key}!").format(
                entry=entry, option_key=option_key
            )
        )
    return num


def positive_integer(entry, option_key="Positive Integer", **kwargs):
    num = signed_integer(entry, option_key)
    if not num >= 1:
        raise ValueError(
            _("Must enter a whole number greater than 0 for {option_key}!").format(
                option_key=option_key
            )
        )
    return num


def unsigned_integer(entry, option_key="Unsigned Integer", **kwargs):
    num = signed_integer(entry, option_key)
    if not num >= 0:
        raise ValueError(
            _("{option_key} must be a whole number greater than " "or equal to 0!").format(
                option_key=option_key
            )
        )
    return num


def boolean(entry, option_key="True/False", **kwargs):
    """
    Simplest check in computer logic, right? This will take user input to flick the switch on or off

    Args:
        entry (str): A value such as True, On, Enabled, Disabled, False, 0, or 1.
        option_key (str): What kind of Boolean we are setting. What Option is this for?

    Returns:
        Boolean

    """
    error = _("Must enter a true/false input for {option_key}. Accepts {alternatives}.").format(
        option_key=option_key, alternatives="0/1, True/False, On/Off, Yes/No, Enabled/Disabled"
    )
    if not isinstance(entry, str):
        raise ValueError(error)
    entry = entry.upper()
    if entry in ("1", "TRUE", "ON", "ENABLED", "ENABLE", "YES"):
        return True
    if entry in ("0", "FALSE", "OFF", "DISABLED", "DISABLE", "NO"):
        return False
    raise ValueError(error)


def timezone(entry, option_key="Timezone", **kwargs):
    """
    Takes user input as string, and partial matches a Timezone.

    Args:
        entry (str): The name of the Timezone.
        option_key (str): What this Timezone is used for.

    Returns:
        A PYTZ timezone.

    """
    if not entry:
        raise ValueError(_("No {option_key} entered!").format(option_key=option_key))
    found = _partial(list(_TZ_DICT.keys()), entry, ret_index=False)
    if len(found) > 1:
        raise ValueError(
            _("That matched: {matches}. Please be more specific!").format(
                matches=", ".join(str(t) for t in found)
            )
        )
    if found:
        return _TZ_DICT[found[0]]
    raise ValueError(
        _("Could not find timezone '{entry}' for {option_key}!").format(
            entry=entry, option_key=option_key
        )
    )


def email(entry, option_key="Email Address", **kwargs):
    if not entry:
        raise ValueError(_("Email address field empty!"))
    valid = validate_email_address(entry)
    if not valid:
        raise ValueError(_("That isn't a valid {option_key}!").format(option_key=option_key))
    return entry


def lock(entry, option_key="locks", access_options=None, **kwargs):
    entry = entry.strip()
    if not entry:
        raise ValueError(_("No {option_key} entered to set!").format(option_key=option_key))
    for locksetting in entry.split(";"):
        access_type, lockfunc = locksetting.split(":", 1)
        if not access_type:
            raise ValueError(_("Must enter an access type!"))
        if access_options:
            if access_type not in access_options:
                raise ValueError(
                    _("Access type must be one of: {alternatives}").format(
                        alternatives=", ".join(access_options)
                    )
                )
        if not lockfunc:
            raise ValueError(_("Lock func not entered."))
    return entry



"""
CmdSethandler

The Cmdsethandler tracks an object's 'Current CmdSet', which is the
current merged sum of all CmdSets added to it.

A CmdSet constitues a set of commands. The CmdSet works as a special
intelligent container that, when added to other CmdSet make sure that
same-name commands are treated correctly (usually so there are no
doublets).  This temporary but up-to-date merger of CmdSet is jointly
called the Current Cmset. It is this Current CmdSet that the
commandhandler looks through whenever an account enters a command (it
also adds CmdSets from objects in the room in real-time). All account
objects have a 'default cmdset' containing all the normal in-game mud
commands (look etc).

So what is all this cmdset complexity good for?

In its simplest form, a CmdSet has no commands, only a key name. In
this case the cmdset's use is up to each individual game - it can be
used by an AI module for example (mobs in cmdset 'roam' move from room
to room, in cmdset 'attack' they enter combat with accounts).

Defining commands in cmdsets offer some further powerful game-design
consequences however. Here are some examples:

As mentioned above, all accounts always have at least the Default
CmdSet.  This contains the set of all normal-use commands in-game,
stuff like look and @desc etc. Now assume our players end up in a dark
room. You don't want the player to be able to do much in that dark
room unless they light a candle. You could handle this by changing all
your normal commands to check if the player is in a dark room. This
rapidly goes unwieldly and error prone. Instead you just define a
cmdset with only those commands you want to be available in the 'dark'
cmdset - maybe a modified look command and a 'light candle' command -
and have this completely replace the default cmdset.

Another example: Say you want your players to be able to go
fishing. You could implement this as a 'fish' command that fails
whenever the account has no fishing rod. Easy enough.  But what if you
want to make fishing more complex - maybe you want four-five different
commands for throwing your line, reeling in, etc? Most players won't
(we assume) have fishing gear, and having all those detailed commands
is cluttering up the command list. And what if you want to use the
'throw' command also for throwing rocks etc instead of 'using it up'
for a minor thing like fishing?

So instead you put all those detailed fishing commands into their own
CommandSet called 'Fishing'. Whenever the player gives the command
'fish' (presumably the code checks there is also water nearby), only
THEN this CommandSet is added to the Cmdhandler of the account. The
'throw' command (which normally throws rocks) is replaced by the
custom 'fishing variant' of throw. What has happened is that the
Fishing CommandSet was merged on top of the Default ones, and due to
how we defined it, its command overrules the default ones.

When we are tired of fishing, we give the 'go home' command (or
whatever) and the Cmdhandler simply removes the fishing CommandSet
so that we are back at defaults (and can throw rocks again).

Since any number of CommandSets can be piled on top of each other, you
can then implement separate sets for different situations. For
example, you can have a 'On a boat' set, onto which you then tack on
the 'Fishing' set. Fishing from a boat? No problem!

"""
import sys
from importlib import import_module
from inspect import trace
from traceback import format_exc

from django.conf import settings
from django.utils.translation import gettext as _

from evennia.commands.cmdset import CmdSet
from evennia.server.models import ServerConfig
from evennia.utils import logger, utils

__all__ = ("import_cmdset", "CmdSetHandler")

_CACHED_CMDSETS = {}
_CMDSET_PATHS = utils.make_iter(settings.CMDSET_PATHS)
_IN_GAME_ERRORS = settings.IN_GAME_ERRORS
_CMDSET_FALLBACKS = settings.CMDSET_FALLBACKS


# Output strings

_ERROR_CMDSET_IMPORT = _(
    """{traceback}
Error loading cmdset '{path}'
(Traceback was logged {timestamp})"""
)

_ERROR_CMDSET_KEYERROR = _(
    """Error loading cmdset: No cmdset class '{classname}' in '{path}'.
(Traceback was logged {timestamp})"""
)

_ERROR_CMDSET_SYNTAXERROR = _(
    """{traceback}
SyntaxError encountered when loading cmdset '{path}'.
(Traceback was logged {timestamp})"""
)

_ERROR_CMDSET_EXCEPTION = _(
    """{traceback}
Compile/Run error when loading cmdset '{path}'.
(Traceback was logged {timestamp})"""
)

_ERROR_CMDSET_FALLBACK = _(
    """
Error encountered for cmdset at path '{path}'.
Replacing with fallback '{fallback_path}'.
"""
)

_ERROR_CMDSET_NO_FALLBACK = _("""Fallback path '{fallback_path}' failed to generate a cmdset.""")


class _ErrorCmdSet(CmdSet):
    """
    This is a special cmdset used to report errors.
    """

    key = "_CMDSET_ERROR"
    errmessage = "Error when loading cmdset."


class _EmptyCmdSet(CmdSet):
    """
    This cmdset represents an empty cmdset
    """

    key = "_EMPTY_CMDSET"
    priority = -101
    mergetype = "Union"


def import_cmdset(path, cmdsetobj, emit_to_obj=None, no_logging=False):
    """
    This helper function is used by the cmdsethandler to load a cmdset
    instance from a python module, given a python_path. It's usually accessed
    through the cmdsethandler's add() and add_default() methods.
    path - This is the full path to the cmdset object on python dot-form

    Args:
        path (str): The path to the command set to load.
        cmdsetobj (CmdSet): The database object/typeclass on which this cmdset is to be
            assigned (this can be also channels and exits, as well as accounts
            but there will always be such an object)
        emit_to_obj (Object, optional): If given, error is emitted to
            this object (in addition to logging)
        no_logging (bool, optional): Don't log/send error messages.
            This can be useful if import_cmdset is just used to check if
            this is a valid python path or not.
    Returns:
        cmdset (CmdSet): The imported command set. If an error was
            encountered, `commands.cmdsethandler._ErrorCmdSet` is returned
            for the benefit of the handler.

    """
    python_paths = [path] + [
        "%s.%s" % (prefix, path) for prefix in _CMDSET_PATHS if not path.startswith(prefix)
    ]
    errstring = ""
    for python_path in python_paths:
        if "." in path:
            modpath, classname = python_path.rsplit(".", 1)
        else:
            raise ImportError(f"The path '{path}' is not on the form modulepath.ClassName")

        try:
            # first try to get from cache
            cmdsetclass = _CACHED_CMDSETS.get(python_path, None)

            if not cmdsetclass:
                try:
                    module = import_module(modpath, package="evennia")
                except ImportError as exc:
                    if len(trace()) > 2:
                        # error in module, make sure to not hide it.
                        dum, dum, tb = sys.exc_info()
                        raise exc.with_traceback(tb)
                    else:
                        # try next suggested path
                        errstring += _("\n(Unsuccessfully tried '{path}').").format(
                            path=python_path
                        )
                        continue
                try:
                    cmdsetclass = getattr(module, classname)
                except AttributeError as exc:
                    if len(trace()) > 2:
                        # Attribute error within module, don't hide it
                        dum, dum, tb = sys.exc_info()
                        raise exc.with_traceback(tb)
                    else:
                        errstring += _("\n(Unsuccessfully tried '{path}').").format(
                            path=python_path
                        )
                        continue
                _CACHED_CMDSETS[python_path] = cmdsetclass

            # instantiate the cmdset (and catch its errors)
            if callable(cmdsetclass):
                cmdsetclass = cmdsetclass(cmdsetobj)
            return cmdsetclass
        except ImportError as err:
            logger.log_trace()
            errstring += _ERROR_CMDSET_IMPORT
            if _IN_GAME_ERRORS:
                errstring = errstring.format(
                    path=python_path, traceback=format_exc(), timestamp=logger.timeformat()
                )
            else:
                errstring = errstring.format(
                    path=python_path, traceback=err, timestamp=logger.timeformat()
                )
            break
        except KeyError:
            logger.log_trace()
            errstring += _ERROR_CMDSET_KEYERROR
            errstring = errstring.format(
                classname=classname, path=python_path, timestamp=logger.timeformat()
            )
            break
        except SyntaxError as err:
            logger.log_trace()
            errstring += _ERROR_CMDSET_SYNTAXERROR
            if _IN_GAME_ERRORS:
                errstring = errstring.format(
                    path=python_path, traceback=format_exc(), timestamp=logger.timeformat()
                )
            else:
                errstring = errstring.format(
                    path=python_path, traceback=err, timestamp=logger.timeformat()
                )
            break
        except Exception as err:
            logger.log_trace()
            errstring += _ERROR_CMDSET_EXCEPTION
            if _IN_GAME_ERRORS:
                errstring = errstring.format(
                    path=python_path, traceback=format_exc(), timestamp=logger.timeformat()
                )
            else:
                errstring = errstring.format(
                    path=python_path, traceback=err, timestamp=logger.timeformat()
                )
            break

    if errstring:
        # returning an empty error cmdset
        errstring = errstring.strip()
        if not no_logging:
            logger.log_err(errstring)
            if emit_to_obj and not ServerConfig.objects.conf("server_starting_mode"):
                emit_to_obj.msg(errstring)
        err_cmdset = _ErrorCmdSet()
        err_cmdset.errmessage = errstring
        return err_cmdset
    return None  # undefined error


# classes


class CmdSetHandler(object):
    """
    The CmdSetHandler is always stored on an object, this object is supplied
    as an argument.

    The 'current' cmdset is the merged set currently active for this object.
    This is the set the game engine will retrieve when determining which
    commands are available to the object. The cmdset_stack holds a history of
    all CmdSets to allow the handler to remove/add cmdsets at will. Doing so
    will re-calculate the 'current' cmdset.
    """

    def __init__(self, obj, init_true=True):
        """
        This method is called whenever an object is recreated.

        Args:
            obj (Object): An reference to the game object this handler
                belongs to.
            init_true (bool, optional): Set when the handler is initializing
                and loads the current cmdset.

        """
        self.obj = obj

        # the id of the "merged" current cmdset for easy access.
        self.key = None
        # this holds the "merged" current command set. Note that while the .update
        # method updates this field in order to have it synced when operating on
        # cmdsets in-code, when the game runs, this field is kept up-to-date by
        # the cmdsethandler's get_and_merge_cmdsets!
        self.current = None
        # this holds a history of CommandSets
        self.cmdset_stack = [_EmptyCmdSet(cmdsetobj=self.obj)]
        # this tracks which mergetypes are actually in play in the stack
        self.mergetype_stack = ["Union"]

        # the subset of the cmdset_paths that are to be stored in the database
        self.persistent_paths = [""]

        if init_true:
            self.update(init_mode=True)  # is then called from the object __init__.

    def __str__(self):
        """
        Display current commands

        """

        strings = ["<CmdSetHandler> stack:"]
        mergelist = []
        if len(self.cmdset_stack) > 1:
            # We have more than one cmdset in stack; list them all
            for snum, cmdset in enumerate(self.cmdset_stack):
                mergelist.append(str(snum + 1))
                strings.append(f" {snum + 1}: {cmdset}")

        # Display the currently active cmdset, limited by self.obj's permissions
        mergetype = self.mergetype_stack[-1]
        if mergetype != self.current.mergetype:
            merged_on = self.cmdset_stack[-2].key
            mergetype = _("custom {mergetype} on cmdset '{cmdset}'")
            mergetype = mergetype.format(mergetype=mergetype, cmdset=merged_on)

        if mergelist:
            # current is a result of mergers
            mergelist = "+".join(mergelist)
            strings.append(f" <Merged {mergelist}>: {self.current}")
        else:
            # current is a single cmdset
            strings.append(" " + str(self.current))
        return "\n".join(strings).rstrip()

    def _import_cmdset(self, cmdset_path, emit_to_obj=None):
        """
        Method wrapper for import_cmdset; Loads a cmdset from a
        module.

        Args:
            cmdset_path (str): The python path to an cmdset object.
            emit_to_obj (Object): The object to send error messages to

        Returns:
            cmdset (Cmdset): The imported cmdset.

        """
        if not emit_to_obj:
            emit_to_obj = self.obj
        return import_cmdset(cmdset_path, self.obj, emit_to_obj)

    def update(self, init_mode=False):
        """
        Re-adds all sets in the handler to have an updated current

        Args:
            init_mode (bool, optional): Used automatically right after
                this handler was created; it imports all persistent cmdsets
                from the database.

        Notes:
            This method is necessary in order to always have a `.current`
            cmdset when working with the cmdsethandler in code. But the
            CmdSetHandler doesn't (cannot) consider external cmdsets and game
            state. This means that the .current calculated from this method
            will likely not match the true current cmdset as determined at
            run-time by `cmdhandler.get_and_merge_cmdsets()`. So in a running
            game the responsibility of keeping `.current` upt-to-date belongs
            to the central `cmdhandler.get_and_merge_cmdsets()`!

        """
        if init_mode:
            # reimport all persistent cmdsets
            storage = self.obj.cmdset_storage
            if storage:
                self.cmdset_stack = []
                for pos, path in enumerate(storage):
                    if pos == 0 and not path:
                        self.cmdset_stack = [_EmptyCmdSet(cmdsetobj=self.obj)]
                    elif path:
                        cmdset = self._import_cmdset(path)
                        if cmdset:
                            if cmdset.key == "_CMDSET_ERROR":
                                # If a cmdset fails to load, check if we have a fallback path to use
                                fallback_path = _CMDSET_FALLBACKS.get(path, None)
                                if fallback_path:
                                    err = _ERROR_CMDSET_FALLBACK.format(
                                        path=path, fallback_path=fallback_path
                                    )
                                    logger.log_err(err)
                                    if _IN_GAME_ERRORS:
                                        self.obj.msg(err)
                                    cmdset = self._import_cmdset(fallback_path)
                                # If no cmdset is returned from the fallback, we can't go further
                                if not cmdset:
                                    err = _ERROR_CMDSET_NO_FALLBACK.format(
                                        fallback_path=fallback_path
                                    )
                                    logger.log_err(err)
                                    if _IN_GAME_ERRORS:
                                        self.obj.msg(err)
                                    continue
                            cmdset.persistent = cmdset.key != "_CMDSET_ERROR"
                            self.cmdset_stack.append(cmdset)

        # merge the stack into a new merged cmdset
        new_current = None
        self.mergetype_stack = []
        for cmdset in self.cmdset_stack:
            try:
                # for cmdset's '+' operator, order matters.
                new_current = cmdset + new_current
            except TypeError:
                continue
            self.mergetype_stack.append(new_current.actual_mergetype)
        self.current = new_current

    def add(self, cmdset, emit_to_obj=None, persistent=False, default_cmdset=False, **kwargs):
        """
        Add a cmdset to the handler, on top of the old ones, unless it
        is set as the default one (it will then end up at the bottom of the stack)

        Args:
            cmdset (CmdSet or str): Can be a cmdset object or the python path
                to such an object.
            emit_to_obj (Object, optional): An object to receive error messages.
            persistent (bool, optional): Let cmdset remain across server reload.
            default_cmdset (Cmdset, optional): Insert this to replace the
                default cmdset position (there is only one such position,
                always at the bottom of the stack).

        Notes:
          An interesting feature of this method is if you were to send
          it an *already instantiated cmdset* (i.e. not a class), the
          current cmdsethandler's obj attribute will then *not* be
          transferred over to this already instantiated set (this is
          because it might be used elsewhere and can cause strange
          effects).  This means you could in principle have the
          handler launch command sets tied to a *different* object
          than the handler. Not sure when this would be useful, but
          it's a 'quirk' that has to be documented.

        """
        if "permanent" in kwargs:
            logger.log_dep("obj.cmdset.add() kwarg 'permanent' has changed name to 'persistent'.")
            persistent = kwargs["permanent"] if persistent is False else persistent

        if not (isinstance(cmdset, str) or utils.inherits_from(cmdset, CmdSet)):
            string = _("Only CmdSets can be added to the cmdsethandler!")
            raise Exception(string)

        if callable(cmdset):
            cmdset = cmdset(self.obj)
        elif isinstance(cmdset, str):
            # this is (maybe) a python path. Try to import from cache.
            cmdset = self._import_cmdset(cmdset)
        if cmdset and cmdset.key != "_CMDSET_ERROR":
            cmdset.persistent = persistent
            if persistent and cmdset.key != "_CMDSET_ERROR":
                # store the path permanently
                storage = self.obj.cmdset_storage or [""]
                if default_cmdset:
                    storage[0] = cmdset.path
                else:
                    storage.append(cmdset.path)
                self.obj.cmdset_storage = storage
            if default_cmdset:
                self.cmdset_stack[0] = cmdset
            else:
                self.cmdset_stack.append(cmdset)
            self.update()

    def add_default(self, cmdset, emit_to_obj=None, persistent=True, **kwargs):
        """
        Shortcut for adding a default cmdset.

        Args:
            cmdset (Cmdset): The Cmdset to add.
            emit_to_obj (Object, optional): Gets error messages
            persistent (bool, optional): The new Cmdset should survive a server reboot.

        """
        if "permanent" in kwargs:
            logger.log_dep(
                "obj.cmdset.add_default() kwarg 'permanent' has changed name to 'persistent'."
            )
            persistent = kwargs["permanent"] if persistent is None else persistent
        self.add(cmdset, emit_to_obj=emit_to_obj, persistent=persistent, default_cmdset=True)

    def remove(self, cmdset=None, default_cmdset=False):
        """
        Remove a cmdset from the  handler.

        Args:
            cmdset (CommandSet or str, optional): This can can be supplied either as a cmdset-key,
                an instance of the CmdSet or a python path to the cmdset.
                If no key is given, the last cmdset in the stack is
                removed. Whenever the cmdset_stack changes, the cmdset is
                updated. If default_cmdset is set, this argument is ignored.
            default_cmdset (bool, optional): If set, this will remove the
                default cmdset (at the bottom of the stack).

        """
        if default_cmdset:
            # remove the default cmdset only
            if self.cmdset_stack:
                cmdset = self.cmdset_stack[0]
                if cmdset.persistent:
                    storage = self.obj.cmdset_storage or [""]
                    storage[0] = ""
                    self.obj.cmdset_storage = storage
                self.cmdset_stack[0] = _EmptyCmdSet(cmdsetobj=self.obj)
            else:
                self.cmdset_stack = [_EmptyCmdSet(cmdsetobj=self.obj)]
            self.update()
            return

        if len(self.cmdset_stack) < 2:
            # don't allow deleting default cmdsets here.
            return

        if not cmdset:
            # remove the last one in the stack
            cmdset = self.cmdset_stack.pop()
            if cmdset.persistent:
                storage = self.obj.cmdset_storage
                storage.pop()
                self.obj.cmdset_storage = storage
        else:
            # try it as a callable
            if callable(cmdset) and hasattr(cmdset, "path"):
                delcmdsets = [cset for cset in self.cmdset_stack[1:] if cset.path == cmdset.path]
            else:
                # try it as a path or key
                delcmdsets = [
                    cset
                    for cset in self.cmdset_stack[1:]
                    if cset.path == cmdset or cset.key == cmdset
                ]
            storage = []

            if any(cset.persistent for cset in delcmdsets):
                # only hit database if there's need to
                storage = self.obj.cmdset_storage
                updated = False
                for cset in delcmdsets:
                    if cset.persistent:
                        try:
                            storage.remove(cset.path)
                            updated = True
                        except ValueError:
                            # nothing to remove
                            pass
                if updated:
                    self.obj.cmdset_storage = storage
            for cset in delcmdsets:
                # clean the in-memory stack
                try:
                    self.cmdset_stack.remove(cset)
                except ValueError:
                    # nothing to remove
                    pass
        # re-sync the cmdsethandler.
        self.update()

    # legacy alias
    delete = remove

    def remove_default(self):
        """
        This explicitly deletes only the default cmdset.

        """
        self.remove(default_cmdset=True)

    # legacy alias
    delete_default = remove_default

    def get(self):
        """
        Get all cmdsets.

        Returns:
            cmdsets (list): All the command sets currently in the handler.

        """
        return self.cmdset_stack

    # backwards-compatible alias
    all = get

    def clear(self):
        """
        Removes all Command Sets from the handler except the default one
        (use `self.remove_default` to remove that).

        """
        self.cmdset_stack = [self.cmdset_stack[0]]
        storage = self.obj.cmdset_storage
        if storage:
            storage = storage[0]
            self.obj.cmdset_storage = storage
        self.update()

    def has(self, cmdset, must_be_default=False):
        """
        checks so the cmdsethandler contains a given cmdset

        Args:
            cmdset (str or Cmdset): Cmdset key, pythonpath or
                Cmdset to check the existence for.
            must_be_default (bool, optional): Only return True if
                the checked cmdset is the default one.

        Returns:
            has_cmdset (bool): Whether or not the cmdset is in the handler.

        """
        if callable(cmdset) and hasattr(cmdset, "path"):
            # try it as a callable
            if must_be_default:
                return self.cmdset_stack and (self.cmdset_stack[0].path == cmdset.path)
            else:
                return any([cset for cset in self.cmdset_stack if cset.path == cmdset.path])
        else:
            # try it as a path or key
            if must_be_default:
                return self.cmdset_stack and (
                    self.cmdset_stack[0].key == cmdset or self.cmdset_stack[0].path == cmdset
                )
            else:
                return any(
                    [
                        cset
                        for cset in self.cmdset_stack
                        if cset.path == cmdset or cset.key == cmdset
                    ]
                )

    # backwards-compatability alias
    has_cmdset = has

    def reset(self):
        """
        Force reload of all cmdsets in handler. This should be called
        after _CACHED_CMDSETS have been cleared (normally this is
        handled automatically by @reload).

        """
        new_cmdset_stack = []
        for cmdset in self.cmdset_stack:
            if cmdset.key == "_EMPTY_CMDSET":
                new_cmdset_stack.append(cmdset)
            else:
                new_cmdset_stack.append(self._import_cmdset(cmdset.path))
        self.cmdset_stack = new_cmdset_stack
        self.update()



# Generated by Django 4.0.2 on 2022-10-31 19:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("scripts", "0015_convert_contrib_paths"),
    ]

    operations = [
        migrations.AlterField(
            model_name="scriptdb",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
    ]



"""
Auditable Server Sessions - Example Outputs
Example methods demonstrating output destinations for logs generated by
audited server sessions.

This is designed to be a single source of events for developers to customize
and add any additional enhancements before events are written out-- i.e. if you
want to keep a running list of what IPs a user logs in from on account/character
objects, or if you want to perform geoip or ASN lookups on IPs before committing,
or tag certain events with the results of a reputational lookup, this should be
the easiest place to do it. Write a method and invoke it via
`settings.AUDIT_CALLBACK` to have log data objects passed to it.

Evennia contribution - Johnny 2017
"""
import json
import syslog

from evennia.utils.logger import log_file


def to_file(data):
    """
    Writes dictionaries of data generated by an AuditedServerSession to files
    in JSON format, bucketed by date.

    Uses Evennia's native logger and writes to the default
    log directory (~/yourgame/server/logs/ or settings.LOG_DIR)

    Args:
        data (dict): Parsed session transmission data.

    """
    # Bucket logs by day and remove objects before serialization
    bucket = data.pop("objects")["time"].strftime("%Y-%m-%d")

    # Write it
    log_file(json.dumps(data), filename="audit_%s.log" % bucket)


def to_syslog(data):
    """
    Writes dictionaries of data generated by an AuditedServerSession to syslog.

    Takes advantage of your system's native logger and writes to wherever
    you have it configured, which is independent of Evennia.
    Linux systems tend to write to /var/log/syslog.

    If you're running rsyslog, you can configure it to dump and/or forward logs
    to disk and/or an external data warehouse (recommended-- if your server is
    compromised or taken down, losing your logs along with it is no help!).

    Args:
        data (dict): Parsed session transmission data.

    """
    # Remove objects before serialization
    data.pop("objects")

    # Write it out
    syslog.syslog(json.dumps(data))



"""
This is the *abstract* django models for many of the database objects
in Evennia. A django abstract (obs, not the same as a Python metaclass!) is
a model which is not actually created in the database, but which only exists
for other models to inherit from, to avoid code duplication. Any model can
import and inherit from these classes.

Attributes are database objects stored on other objects. The implementing
class needs to supply a ForeignKey field attr_object pointing to the kind
of object being mapped. Attributes storing iterables actually store special
types of iterables named PackedList/PackedDict respectively. These make
sure to save changes to them to database - this is criticial in order to
allow for obj.db.mylist[2] = data. Also, all dbobjects are saved as
dbrefs but are also aggressively cached.

TypedObjects are objects 'decorated' with a typeclass - that is, the typeclass
(which is a normal Python class implementing some special tricks with its
get/set attribute methods, allows for the creation of all sorts of different
objects all with the same database object underneath. Usually attributes are
used to permanently store things not hard-coded as field on the database object.
The admin should usually not have to deal directly  with the database object
layer.

This module also contains the Managers for the respective models; inherit from
these to create custom managers.

"""
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import signals
from django.db.models.base import ModelBase
from django.urls import reverse
from django.utils.encoding import smart_str
from django.utils.text import slugify

import evennia
from evennia.locks.lockhandler import LockHandler
from evennia.server.signals import SIGNAL_TYPED_OBJECT_POST_RENAME
from evennia.typeclasses import managers
from evennia.typeclasses.attributes import (
    Attribute,
    AttributeHandler,
    AttributeProperty,
    DbHolder,
    InMemoryAttributeBackend,
    ModelAttributeBackend,
)
from evennia.typeclasses.tags import (
    AliasHandler,
    PermissionHandler,
    Tag,
    TagCategoryProperty,
    TagHandler,
    TagProperty,
)
from evennia.utils.idmapper.models import SharedMemoryModel, SharedMemoryModelBase
from evennia.utils.logger import log_trace
from evennia.utils.utils import class_from_module, inherits_from, is_iter, lazy_property

__all__ = ("TypedObject",)

TICKER_HANDLER = None

_PERMISSION_HIERARCHY = [p.lower() for p in settings.PERMISSION_HIERARCHY]
_TYPECLASS_AGGRESSIVE_CACHE = settings.TYPECLASS_AGGRESSIVE_CACHE
_GA = object.__getattribute__
_SA = object.__setattr__


# signal receivers. Connected in __new__


def call_at_first_save(sender, instance, created, **kwargs):
    """
    Receives a signal just after the object is saved.

    """
    if created:
        instance.at_first_save()


def remove_attributes_on_delete(sender, instance, **kwargs):
    """
    Wipe object's Attributes when it's deleted

    """
    instance.db_attributes.all().delete()


# ------------------------------------------------------------
#
# Typed Objects
#
# ------------------------------------------------------------


#
# Meta class for typeclasses
#


class TypeclassBase(SharedMemoryModelBase):
    """
    Metaclass which should be set for the root of model proxies
    that don't define any new fields, like Object, Script etc. This
    is the basis for the typeclassing system.

    """

    def __new__(cls, name, bases, attrs):
        """
        We must define our Typeclasses as proxies. We also store the
        path directly on the class, this is required by managers.
        """

        # storage of stats
        attrs["typename"] = name
        attrs["path"] = "%s.%s" % (attrs["__module__"], name)

        def _get_dbmodel(bases):
            """Recursively get the dbmodel"""
            if not hasattr(bases, "__iter__"):
                bases = [bases]
            for base in bases:
                try:
                    if base._meta.proxy or base._meta.abstract:
                        for kls in base._meta.parents:
                            return _get_dbmodel(kls)
                except AttributeError:
                    # this happens if trying to parse a non-typeclass mixin parent,
                    # without a _meta
                    continue
                else:
                    return base
                return None

        dbmodel = _get_dbmodel(bases)

        if not dbmodel:
            raise TypeError(f"{name} does not appear to inherit from a database model.")

        # typeclass proxy setup
        # first check explicit __applabel__ on the typeclass, then figure
        # it out from the dbmodel
        if "__applabel__" not in attrs:
            # find the app-label in one of the bases, usually the dbmodel
            attrs["__applabel__"] = dbmodel._meta.app_label

        if "Meta" not in attrs:

            class Meta:
                proxy = True
                app_label = attrs.get("__applabel__", "typeclasses")

            attrs["Meta"] = Meta
        attrs["Meta"].proxy = True

        new_class = ModelBase.__new__(cls, name, bases, attrs)

        # django doesn't support inheriting proxy models so we hack support for
        # it here by injecting `proxy_for_model` to the actual dbmodel.
        # Unfortunately we cannot also set the correct model_name, because this
        # would block multiple-inheritance of typeclasses (Django doesn't allow
        # multiple bases of the same model).
        if dbmodel:
            new_class._meta.proxy_for_model = dbmodel
            # Maybe Django will eventually handle this in the future:
            # new_class._meta.model_name = dbmodel._meta.model_name

        # attach signals
        signals.post_save.connect(call_at_first_save, sender=new_class)
        signals.pre_delete.connect(remove_attributes_on_delete, sender=new_class)
        return new_class


#
# Main TypedObject abstraction
#


class TypedObject(SharedMemoryModel):
    """
    Abstract Django model.

    This is the basis for a typed object. It also contains all the
    mechanics for managing connected attributes.

    The TypedObject has the following properties:

    - key - main name
    - name - alias for key
    - typeclass_path - the path to the decorating typeclass
    - typeclass - auto-linked typeclass
    - date_created - time stamp of object creation
    - permissions - perm strings
    - dbref - #id of object
    - db - persistent attribute storage
    - ndb - non-persistent attribute storage

    """

    #
    # TypedObject Database Model setup
    #
    #
    # These databse fields are all accessed and set using their corresponding
    # properties, named same as the field, but without the db_* prefix
    # (no separate save() call is needed)

    # Main identifier of the object, for searching. Is accessed with self.key
    # or self.name
    db_key = models.CharField("key", max_length=255, db_index=True)
    # This is the python path to the type class this object is tied to. The
    # typeclass is what defines what kind of Object this is)
    db_typeclass_path = models.CharField(
        "typeclass",
        max_length=255,
        null=True,
        help_text=(
            "this defines what 'type' of entity this is. This variable holds "
            "a Python path to a module with a valid Evennia Typeclass."
        ),
        db_index=True,
    )
    # Creation date. This is not changed once the object is created.
    db_date_created = models.DateTimeField("creation date", editable=False, auto_now_add=True)
    # Lock storage
    db_lock_storage = models.TextField(
        "locks",
        blank=True,
        help_text=(
            "locks limit access to an entity. A lock is defined as a 'lock string' "
            "on the form 'type:lockfunctions', defining what functionality is locked and "
            "how to determine access. Not defining a lock means no access is granted."
        ),
    )
    # many2many relationships
    db_attributes = models.ManyToManyField(
        Attribute,
        help_text=(
            "attributes on this object. An attribute can hold any pickle-able "
            "python object (see docs for special cases)."
        ),
    )
    db_tags = models.ManyToManyField(
        Tag,
        help_text=(
            "tags on this object. Tags are simple string markers to identify, "
            "group and alias objects."
        ),
    )

    # Database manager
    objects = managers.TypedObjectManager()

    # quick on-object typeclass cache for speed
    _cached_typeclass = None

    # typeclass mechanism

    def set_class_from_typeclass(self, typeclass_path=None):
        if typeclass_path:
            try:
                self.__class__ = class_from_module(
                    typeclass_path, defaultpaths=settings.TYPECLASS_PATHS
                )
            except Exception:
                log_trace()
                try:
                    self.__class__ = class_from_module(self.__settingsclasspath__)
                except Exception:
                    log_trace()
                    try:
                        self.__class__ = class_from_module(self.__defaultclasspath__)
                    except Exception:
                        log_trace()
                        self.__class__ = self._meta.concrete_model or self.__class__
            finally:
                self.db_typeclass_path = typeclass_path
        elif self.db_typeclass_path:
            try:
                self.__class__ = class_from_module(self.db_typeclass_path)
            except Exception:
                log_trace()
                try:
                    self.__class__ = class_from_module(self.__defaultclasspath__)
                except Exception:
                    log_trace()
                    self.__dbclass__ = self._meta.concrete_model or self.__class__
        else:
            self.db_typeclass_path = "%s.%s" % (self.__module__, self.__class__.__name__)
        # important to put this at the end since _meta is based on the set __class__
        try:
            self.__dbclass__ = self._meta.concrete_model or self.__class__
        except AttributeError:
            err_class = repr(self.__class__)
            self.__class__ = class_from_module("evennia.objects.objects.DefaultObject")
            self.__dbclass__ = class_from_module("evennia.objects.models.ObjectDB")
            self.db_typeclass_path = "evennia.objects.objects.DefaultObject"
            log_trace(
                "Critical: Class %s of %s is not a valid typeclass!\nTemporarily falling back"
                " to %s." % (err_class, self, self.__class__)
            )

    def __init__(self, *args, **kwargs):
        """
        The `__init__` method of typeclasses is the core operational
        code of the typeclass system, where it dynamically re-applies
        a class based on the db_typeclass_path database field rather
        than use the one in the model.

        Args:
            Passed through to parent.

        Keyword Args:
            Passed through to parent.

        Notes:
            The loading mechanism will attempt the following steps:

            1. Attempt to load typeclass given on command line
            2. Attempt to load typeclass stored in db_typeclass_path
            3. Attempt to load `__settingsclasspath__`, which is by the
               default classes defined to be the respective user-set
               base typeclass settings, like `BASE_OBJECT_TYPECLASS`.
            4. Attempt to load `__defaultclasspath__`, which is the
               base classes in the library, like DefaultObject etc.
            5. If everything else fails, use the database model.

            Normal operation is to load successfully at either step 1
            or 2 depending on how the class was called. Tracebacks
            will be logged for every step the loader must take beyond
            2.

        """
        typeclass_path = kwargs.pop("typeclass", None)
        super().__init__(*args, **kwargs)
        self.set_class_from_typeclass(typeclass_path=typeclass_path)

    def init_evennia_properties(self):
        """
        Called by creation methods; makes sure to initialize Attribute/TagProperties
        by fetching them once.
        """
        for propkey, prop in self.__class__.__dict__.items():
            if isinstance(prop, (AttributeProperty, TagProperty, TagCategoryProperty)):
                try:
                    getattr(self, propkey)
                except Exception:
                    log_trace()

    # initialize all handlers in a lazy fashion
    @lazy_property
    def attributes(self):
        return AttributeHandler(self, ModelAttributeBackend)

    @lazy_property
    def locks(self):
        return LockHandler(self)

    @lazy_property
    def tags(self):
        return TagHandler(self)

    @lazy_property
    def aliases(self):
        return AliasHandler(self)

    @lazy_property
    def permissions(self):
        return PermissionHandler(self)

    @lazy_property
    def nattributes(self):
        return AttributeHandler(self, InMemoryAttributeBackend)

    class Meta:
        """
        Django setup info.
        """

        abstract = True
        verbose_name = "Evennia Database Object"
        ordering = ["-db_date_created", "id", "db_typeclass_path", "db_key"]

    # wrapper
    # Wrapper properties to easily set database fields. These are
    # @property decorators that allows to access these fields using
    # normal python operations (without having to remember to save()
    # etc). So e.g. a property 'attr' has a get/set/del decorator
    # defined that allows the user to do self.attr = value,
    # value = self.attr and del self.attr respectively (where self
    # is the object in question).

    # name property (alias to self.key)
    def __name_get(self):
        return self.key

    def __name_set(self, value):
        self.key = value

    def __name_del(self):
        raise Exception("Cannot delete name")

    name = property(__name_get, __name_set, __name_del)

    # key property (overrides's the idmapper's db_key for the at_rename hook)
    @property
    def key(self):
        return self.db_key

    @key.setter
    def key(self, value):
        oldname = str(self.db_key)
        self.db_key = value
        self.save(update_fields=["db_key"])
        self.at_rename(oldname, value)
        SIGNAL_TYPED_OBJECT_POST_RENAME.send(sender=self, old_key=oldname, new_key=value)

    #
    #
    # TypedObject main class methods and properties
    #
    #

    def __eq__(self, other):
        try:
            return self.__dbclass__ == other.__dbclass__ and self.dbid == other.dbid
        except AttributeError:
            return False

    def __hash__(self):
        # this is required to maintain hashing
        return super().__hash__()

    def __str__(self):
        return smart_str("%s" % self.db_key)

    def __repr__(self):
        return "%s" % self.db_key

    # @property
    def __dbid_get(self):
        """
        Caches and returns the unique id of the object.
        Use this instead of self.id, which is not cached.
        """
        return self.id

    def __dbid_set(self, value):
        raise Exception("dbid cannot be set!")

    def __dbid_del(self):
        raise Exception("dbid cannot be deleted!")

    dbid = property(__dbid_get, __dbid_set, __dbid_del)

    # @property
    def __dbref_get(self):
        """
        Returns the object's dbref on the form #NN.
        """
        return "#%s" % self.id

    def __dbref_set(self):
        raise Exception("dbref cannot be set!")

    def __dbref_del(self):
        raise Exception("dbref cannot be deleted!")

    dbref = property(__dbref_get, __dbref_set, __dbref_del)

    def at_idmapper_flush(self):
        """
        This is called when the idmapper cache is flushed and
        allows customized actions when this happens.

        Returns:
            do_flush (bool): If True, flush this object as normal. If
                False, don't flush and expect this object to handle
                the flushing on its own.

        Notes:
            The default implementation relies on being able to clear
            Django's Foreignkey cache on objects not affected by the
            flush (notably objects with an NAttribute stored). We rely
            on this cache being stored on the format "_<fieldname>_cache".
            If Django were to change this name internally, we need to
            update here (unlikely, but marking just in case).

        """
        if self.nattributes.all():
            # we can't flush this object if we have non-persistent
            # attributes stored - those would get lost! Nevertheless
            # we try to flush as many references as we can.
            self.attributes.reset_cache()
            self.tags.reset_cache()
            # flush caches for all related fields
            for field in self._meta.fields:
                name = "_%s_cache" % field.name
                if field.is_relation and name in self.__dict__:
                    # a foreignkey - remove its cache
                    del self.__dict__[name]
            return False
        # a normal flush
        return True

    #
    # Object manipulation methods
    #

    def at_init(self):
        """
        Called when this object is loaded into cache. This is  more reliable
        than to override `__init__`.

        """
        pass

    @classmethod
    def search(cls, query, **kwargs):
        """
        Overridden by class children. This implements a common API.

        Args:
            query (str): A search query.
            **kwargs: Other search parameters.

        Returns:
            list: A list of 0, 1 or more matches, only of this typeclass.

        """
        if cls.objects.dbref(query):
            return [cls.objects.get_id(query)]
        return list(cls.objects.filter(db_key__lower=query))

    def is_typeclass(self, typeclass, exact=False):
        """
        Returns true if this object has this type OR has a typeclass
        which is an subclass of the given typeclass. This operates on
        the actually loaded typeclass (this is important since a
        failing typeclass may instead have its default currently
        loaded) typeclass - can be a class object or the python path
        to such an object to match against.

        Args:
            typeclass (str or class): A class or the full python path
                to the class to check.
            exact (bool, optional): Returns true only if the object's
                type is exactly this typeclass, ignoring parents.

        Returns:
            is_typeclass (bool): If this typeclass matches the given
                typeclass.

        """
        if isinstance(typeclass, str):
            typeclass = [typeclass] + [
                "%s.%s" % (prefix, typeclass) for prefix in settings.TYPECLASS_PATHS
            ]
        else:
            typeclass = [typeclass.path]

        selfpath = self.path
        if exact:
            # check only exact match
            return selfpath in typeclass
        else:
            # check parent chain
            return any(
                hasattr(cls, "path") and cls.path in typeclass for cls in self.__class__.mro()
            )

    def swap_typeclass(
        self,
        new_typeclass,
        clean_attributes=False,
        run_start_hooks="all",
        no_default=True,
        clean_cmdsets=False,
    ):
        """
        This performs an in-situ swap of the typeclass. This means
        that in-game, this object will suddenly be something else.
        Account will not be affected. To 'move' an account to a different
        object entirely (while retaining this object's type), use
        self.account.swap_object().

        Note that this might be an error prone operation if the
        old/new typeclass was heavily customized - your code
        might expect one and not the other, so be careful to
        bug test your code if using this feature! Often its easiest
        to create a new object and just swap the account over to
        that one instead.

        Args:
            new_typeclass (str or classobj): Type to switch to.
            clean_attributes (bool or list, optional): Will delete all
                attributes stored on this object (but not any of the
                database fields such as name or location). You can't get
                attributes back, but this is often the safest bet to make
                sure nothing in the new typeclass clashes with the old
                one. If you supply a list, only those named attributes
                will be cleared.
            run_start_hooks (str or None, optional): This is either None,
                to not run any hooks, "all" to run all hooks defined by
                at_first_start, or a string with space-separated hook-names to run
                (for example 'at_object_creation'). This will
                always be called without arguments.
            no_default (bool, optiona): If set, the swapper will not
                allow for swapping to a default typeclass in case the
                given one fails for some reason. Instead the old one will
                be preserved.
            clean_cmdsets (bool, optional): Delete all cmdsets on the object.

        """

        if not callable(new_typeclass):
            # this is an actual class object - build the path
            new_typeclass = class_from_module(new_typeclass, defaultpaths=settings.TYPECLASS_PATHS)

        # if we get to this point, the class is ok.

        if inherits_from(self, "evennia.scripts.models.ScriptDB"):
            if self.interval > 0:
                raise RuntimeError(
                    "Cannot use swap_typeclass on time-dependent "
                    "Script '%s'.\nStop and start a new Script of the "
                    "right type instead." % self.key
                )

        self.typeclass_path = new_typeclass.path
        self.__class__ = new_typeclass

        if clean_attributes:
            # Clean out old attributes
            if is_iter(clean_attributes):
                for attr in clean_attributes:
                    self.attributes.remove(attr)
                for nattr in clean_attributes:
                    if hasattr(self.ndb, nattr):
                        self.nattributes.remove(nattr)
            else:
                self.attributes.clear()
                self.nattributes.clear()
        if clean_cmdsets:
            # purge all cmdsets
            self.cmdset.clear()
            self.cmdset.remove_default()

        if run_start_hooks == "all":
            # fake this call to mimic the first save
            self.at_first_save()
        elif run_start_hooks:
            # a custom hook-name to call.
            for start_hook in str(run_start_hooks).split():
                getattr(self, run_start_hooks)()

    #
    # Lock / permission methods
    #

    def access(
        self, accessing_obj, access_type="read", default=False, no_superuser_bypass=False, **kwargs
    ):
        """
        Determines if another object has permission to access this one.

        Args:
            accessing_obj (str): Object trying to access this one.
            access_type (str, optional): Type of access sought.
            default (bool, optional): What to return if no lock of
                access_type was found
            no_superuser_bypass (bool, optional): Turn off the
                superuser lock bypass (be careful with this one).

        Keyword Args:
            kwar (any): Ignored, but is there to make the api
                consistent with the object-typeclass method access, which
                use it to feed to its hook methods.

        """
        return self.locks.check(
            accessing_obj,
            access_type=access_type,
            default=default,
            no_superuser_bypass=no_superuser_bypass,
        )

    def check_permstring(self, permstring):
        """
        This explicitly checks if we hold particular permission
        without involving any locks.

        Args:
            permstring (str): The permission string to check against.

        Returns:
            result (bool): If the permstring is passed or not.

        """
        if inherits_from(self, evennia.DefaultObject):
            if (
                self.account
                and self.account.is_superuser
                and not self.account.attributes.get("_quell")
            ):
                return True
        else:
            if self.is_superuser and not self.attributes.get("_quell"):
                return True

        if not permstring:
            return False
        perm = permstring.lower()
        perms = [p.lower() for p in self.permissions.all()]
        if perm in perms:
            # simplest case - we have a direct match
            return True
        if perm in _PERMISSION_HIERARCHY:
            # check if we have a higher hierarchy position
            ppos = _PERMISSION_HIERARCHY.index(perm)
            return any(
                True
                for hpos, hperm in enumerate(_PERMISSION_HIERARCHY)
                if hperm in perms and hpos > ppos
            )
        # we ignore pluralization (english only)
        if perm.endswith("s"):
            return self.check_permstring(perm[:-1])

        return False

    #
    # Deletion methods
    #

    def _deleted(self, *args, **kwargs):
        """
        Scrambling method for already deleted objects
        """
        raise ObjectDoesNotExist("This object was already deleted!")

    def delete(self):
        """
        Cleaning up handlers on the typeclass level

        """
        global TICKER_HANDLER
        self.permissions.clear()
        self.attributes.clear()
        self.aliases.clear()
        if hasattr(self, "nicks"):
            self.nicks.clear()
        # scrambling properties
        self.delete = self._deleted
        super().delete()

    #
    # Attribute storage
    #

    @property
    def db(self):
        """
        Attribute handler wrapper. Allows for the syntax

        ```python
           obj.db.attrname = value
           # and
           value = obj.db.attrname
           # and
           del obj.db.attrname
           # and
           all_attr = obj.db.all()
           # (unless there is an attribute
           #  named 'all', in which case that will be returned instead).
        ```

        """
        try:
            return self._db_holder
        except AttributeError:
            self._db_holder = DbHolder(self, "attributes")
            return self._db_holder

    @db.setter
    def db(self, value):
        "Stop accidentally replacing the db object"
        string = "Cannot assign directly to db object! "
        string += "Use db.attr=value instead."
        raise Exception(string)

    @db.deleter
    def db(self):
        "Stop accidental deletion."
        raise Exception("Cannot delete the db object!")

    #
    # Non-persistent (ndb) storage
    #

    @property
    def ndb(self):
        """
        A non-attr_obj store (ndb: NonDataBase). Everything stored
        to this is guaranteed to be cleared when a server is shutdown.
        Syntax is same as for the _get_db_holder() method and
        property, e.g. obj.ndb.attr = value etc.
        """
        try:
            return self._ndb_holder
        except AttributeError:
            self._ndb_holder = DbHolder(self, "nattrhandler", manager_name="nattributes")
            return self._ndb_holder

    @ndb.setter
    def ndb(self, value):
        "Stop accidentally replacing the ndb object"
        string = "Cannot assign directly to ndb object! "
        string += "Use ndb.attr=value instead."
        raise Exception(string)

    @ndb.deleter
    def ndb(self):
        "Stop accidental deletion."
        raise Exception("Cannot delete the ndb object!")

    def get_display_name(self, looker, **kwargs):
        """
        Displays the name of the object in a viewer-aware manner.

        Args:
            looker (TypedObject, optional): The object or account that is looking
                at/getting inforamtion for this object. If not given, some
                'safe' minimum level should be returned.

        Returns:
            name (str): A string containing the name of the object,
                including the DBREF if this user is privileged to control
                said object.

        Notes:
            This function could be extended to change how object names
            appear to users in character, but be wary. This function
            does not change an object's keys or aliases when
            searching, and is expected to produce something useful for
            builders.

        """
        if self.access(looker, access_type="controls"):
            return "{}(#{})".format(self.name, self.id)
        return self.name

    def get_extra_info(self, looker, **kwargs):
        """
        Used when an object is in a list of ambiguous objects as an
        additional information tag.

        For instance, if you had potions which could have varying
        levels of liquid left in them, you might want to display how
        many drinks are left in each when selecting which to drop, but
        not in your normal inventory listing.

        Args:
            looker (TypedObject): The object or account that is looking
                at/getting information for this object.

        Returns:
            info (str): A string with disambiguating information,
                conventionally with a leading space.

        """

        if self.location == looker:
            return " (carried)"
        return ""

    def at_rename(self, oldname, newname):
        """
        This Hook is called by @name on a successful rename.

        Args:
            oldname (str): The instance's original name.
            newname (str): The new name for the instance.

        """
        pass

    #
    # Web/Django methods
    #

    def web_get_admin_url(self):
        """
        Returns the URI path for the Django Admin page for this object.

        ex. Account#1 = '/admin/accounts/accountdb/1/change/'

        Returns:
            path (str): URI path to Django Admin page for object.

        """
        content_type = ContentType.objects.get_for_model(self.__class__)
        return reverse(
            "admin:%s_%s_change" % (content_type.app_label, content_type.model), args=(self.id,)
        )

    @classmethod
    def web_get_create_url(cls):
        """
        Returns the URI path for a View that allows users to create new
        instances of this object.

        ex. Chargen = '/characters/create/'

        For this to work, the developer must have defined a named view somewhere
        in urls.py that follows the format 'modelname-action', so in this case
        a named view of 'character-create' would be referenced by this method.

        ex.
        url(r'characters/create/', ChargenView.as_view(), name='character-create')

        If no View has been created and defined in urls.py, returns an
        HTML anchor.

        This method is naive and simply returns a path. Securing access to
        the actual view and limiting who can create new objects is the
        developer's responsibility.

        Returns:
            path (str): URI path to object creation page, if defined.

        """
        try:
            return reverse("%s-create" % slugify(cls._meta.verbose_name))
        except Exception:
            return "#"

    def web_get_detail_url(self):
        """
        Returns the URI path for a View that allows users to view details for
        this object.

        Returns:
            path (str): URI path to object detail page, if defined.

        Examples:

            ```python
            Oscar (Character) = '/characters/oscar/1/'
            ```

            For this to work, the developer must have defined a named view somewhere
            in urls.py that follows the format 'modelname-action', so in this case
            a named view of 'character-detail' would be referenced by this method.


            ```python
            url(r'characters/(?P<slug>[\w\d\-]+)/(?P<pk>[0-9]+)/$',
                CharDetailView.as_view(), name='character-detail')
            ```

            If no View has been created and defined in urls.py, returns an
            HTML anchor.

            This method is naive and simply returns a path. Securing access to
            the actual view and limiting who can view this object is the
            developer's responsibility.

        """
        try:
            return reverse(
                "%s-detail" % slugify(self._meta.verbose_name),
                kwargs={"pk": self.pk, "slug": slugify(self.name)},
            )
        except Exception:
            return "#"

    def web_get_puppet_url(self):
        """
        Returns the URI path for a View that allows users to puppet a specific
        object.

        Returns:
            str: URI path to object puppet page, if defined.

        Examples:
            ::

                Oscar (Character) = '/characters/oscar/1/puppet/'

            For this to work, the developer must have defined a named view somewhere
            in urls.py that follows the format 'modelname-action', so in this case
            a named view of 'character-puppet' would be referenced by this method.
            ::

                url(r'characters/(?P<slug>[\w\d\-]+)/(?P<pk>[0-9]+)/puppet/$',
                    CharPuppetView.as_view(), name='character-puppet')

            If no View has been created and defined in urls.py, returns an
            HTML anchor.

            This method is naive and simply returns a path. Securing access to
            the actual view and limiting who can view this object is the developer's
            responsibility.


        """
        try:
            return reverse(
                "%s-puppet" % slugify(self._meta.verbose_name),
                kwargs={"pk": self.pk, "slug": slugify(self.name)},
            )
        except Exception:
            return "#"

    def web_get_update_url(self):
        """
        Returns the URI path for a View that allows users to update this
        object.

        Returns:
            str: URI path to object update page, if defined.

        Examples:

            ```python
            Oscar (Character) = '/characters/oscar/1/change/'
            ```

            For this to work, the developer must have defined a named view somewhere
            in urls.py that follows the format 'modelname-action', so in this case
            a named view of 'character-update' would be referenced by this method.
            ::

                url(r'characters/(?P<slug>[\w\d\-]+)/(?P<pk>[0-9]+)/change/$',
                CharUpdateView.as_view(), name='character-update')

            If no View has been created and defined in urls.py, returns an
            HTML anchor.

            This method is naive and simply returns a path. Securing access to
            the actual view and limiting who can modify objects is the developer's
            responsibility.


        """
        try:
            return reverse(
                "%s-update" % slugify(self._meta.verbose_name),
                kwargs={"pk": self.pk, "slug": slugify(self.name)},
            )
        except Exception:
            return "#"

    def web_get_delete_url(self):
        """
        Returns the URI path for a View that allows users to delete this object.

        Returns:
            path (str): URI path to object deletion page, if defined.

        Examples:

            ```python
            Oscar (Character) = '/characters/oscar/1/delete/'
            ```

            For this to work, the developer must have defined a named view
            somewhere in urls.py that follows the format 'modelname-action', so
            in this case a named view of 'character-detail' would be referenced
            by this method.
            ::

                url(r'characters/(?P<slug>[\w\d\-]+)/(?P<pk>[0-9]+)/delete/$',
                CharDeleteView.as_view(), name='character-delete')

            If no View has been created and defined in urls.py, returns an HTML
            anchor.

            This method is naive and simply returns a path. Securing access to
            the actual view and limiting who can delete this object is the
            developer's responsibility.


        """
        try:
            return reverse(
                "%s-delete" % slugify(self._meta.verbose_name),
                kwargs={"pk": self.pk, "slug": slugify(self.name)},
            )
        except Exception:
            return "#"

    # Used by Django Sites/Admin
    get_absolute_url = web_get_detail_url
